from pathlib import Path

import pytest

from sbomgrader.core.definitions import COOKBOOKS_DIR
from sbomgrader.core.utils import get_mapping
from sbomgrader.grade.cookbooks import Cookbook
from sbomgrader.grade.rules import Document


@pytest.fixture()
def testdata_dir() -> Path:
    return Path(__file__).parent / "testdata"


@pytest.fixture()
def image_index_build_sbom(testdata_dir) -> Document:
    return Document(get_mapping(testdata_dir / "image_index_build_sbom.json"))


@pytest.fixture()
def image_index_release_sbom(testdata_dir) -> Document:
    return Document(get_mapping(testdata_dir / "image_index_release_sbom.json"))


@pytest.fixture()
def image_build_sbom(testdata_dir) -> Document:
    return Document(get_mapping(testdata_dir / "image_build_sbom.json"))


@pytest.fixture()
def image_release_sbom(testdata_dir) -> Document:
    return Document(get_mapping(testdata_dir / "image_release_sbom.json"))


@pytest.fixture()
def product_sbom(testdata_dir) -> Document:
    return Document(get_mapping(testdata_dir / "product_sbom.json"))


@pytest.fixture()
def rpm_build_sbom(testdata_dir) -> Document:
    return Document(get_mapping(testdata_dir / "rpm_build_sbom.json"))


@pytest.fixture()
def rpm_release_sbom(testdata_dir) -> Document:
    return Document(get_mapping(testdata_dir / "rpm_release_sbom.json"))


@pytest.fixture()
def cookbooks_dir() -> Path:
    return COOKBOOKS_DIR


@pytest.fixture()
def image_build_cookbook(cookbooks_dir) -> Cookbook:
    return Cookbook.from_file(cookbooks_dir / "image_build.yml")


@pytest.fixture()
def image_index_build_cookbook(cookbooks_dir) -> Cookbook:
    return Cookbook.from_file(cookbooks_dir / "image_index_build.yml")


@pytest.fixture()
def image_index_release_cookbook(cookbooks_dir) -> Cookbook:
    return Cookbook.from_file(cookbooks_dir / "image_index_release.yml")


@pytest.fixture()
def image_release_cookbook(cookbooks_dir) -> Cookbook:
    return Cookbook.from_file(cookbooks_dir / "image_release.yml")


@pytest.fixture()
def product_cookbook(cookbooks_dir) -> Cookbook:
    return Cookbook.from_file(cookbooks_dir / "product.yml")


@pytest.fixture()
def rpm_build_cookbook(cookbooks_dir) -> Cookbook:
    return Cookbook.from_file(cookbooks_dir / "rpm_build.yml")


@pytest.fixture()
def rpm_release_cookbook(cookbooks_dir) -> Cookbook:
    return Cookbook.from_file(cookbooks_dir / "rpm_release.yml")



import sys
from pathlib import Path

from runpy import run_path
from typing import Callable, Union


class PythonLoader:
    def __init__(self, *file_references: str | Path):
        self._unloaded_file_references: set[Path] = set()
        self._loaded_file_references: set[Path] = set()
        self.__functions: dict[str, Callable] = {}

        self.add_file_references(*file_references)

    def add_file_references(self, *file_references: str | Path) -> None:
        self._unloaded_file_references.update(
            ref for ref in file_references if isinstance(ref, Path)
        )
        self._unloaded_file_references.update(
            Path(ref) for ref in file_references if isinstance(ref, str)
        )

    def _load_all_references(self) -> None:
        for ref in self._unloaded_file_references:
            if not ref.exists():
                continue
            module = run_path(str(ref.absolute()))
            self.__functions.update(
                {name: value for name, value in module.items() if callable(value)}
            )
        self._loaded_file_references.update(self._unloaded_file_references)
        self._unloaded_file_references = set()

    @property
    def file_references(self) -> set[Path]:
        return self._unloaded_file_references | self._loaded_file_references

    def load_func(self, name: str) -> Union[Callable, None]:
        if self._unloaded_file_references:
            self._load_all_references()
        if name not in self.__functions:
            print(
                f"Could not load transformer {name} from files {self.file_references}.",
                file=sys.stderr,
            )
        return self.__functions.get(name, None)



SPDX_CDX_HASHES = {
    # Sources: https://spdx.github.io/spdx-spec/v2.3/package-information/
    # https://cyclonedx.org/guides/OWASP_CycloneDX-Authoritative-Guide-to-SBOM-en.pdf
    ("SHA1", "SHA-1"),
    ("SHA224", "SHA-224"),  # This does not really exist in CDX
    ("SHA256", "SHA-256"),
    ("SHA384", "SHA-384"),
    ("SHA512", "SHA-512"),
    ("SHA3-256", "SHA-3"),
    ("SHA3-384", "SHA-3"),
    ("SHA3-512", "SHA-3"),
    ("BLAKE2b-256", "BLAKE2b"),
    ("BLAKE2b-384", "BLAKE2b"),
    ("BLAKE2b-512", "BLAKE2b"),
    ("BLAKE3", "BLAKE3"),
    ("MD2", "MD2"),  # Not
    ("MD4", "MD4"),  # supported
    ("MD5", "MD5"),  # in
    ("MD6", "MD6"),  # CDX
    ("ADLER32", "ADLER32"),  # either
}
SPDX_CDX_RELATIONSHIPS = {("PACKAGE_OF", "provides"), ("CONTAINS", "")}



import json
from typing import Any

import pytest

from sbomgrader.core.documents import Document
from sbomgrader.core.formats import SBOMFormat
from sbomgrader.translate.prune import prune
from sbomgrader.translate.translation_map import TranslationMap
from sbomgrader.core.utils import get_mapping


@pytest.fixture(scope="session")
def built_in_translation_map() -> TranslationMap:
    yield TranslationMap.from_file(
        "sbomgrader/translation_maps/red_hat_spdx23_cdx16.yml"
    )


def ordered(obj: Any):
    """Orders lists so translation is not order-dependent."""
    if isinstance(obj, dict):
        return sorted((k, ordered(v)) for k, v in obj.items())
    if isinstance(obj, list):
        return sorted(ordered(x) for x in obj)
    else:
        return obj


class ComparableDoc:
    def __init__(self, doc: Document):
        self.document = doc

    def __eq__(self, other):
        if not isinstance(other, ComparableDoc) or not isinstance(other, Document):
            raise TypeError(
                f"Cannot compare object of type {self.__class__.__name__} to an object  of type {other.__class__.__name__}."
            )


@pytest.mark.parametrize(["spdx"], [("spdx23",), ("spdx22",)])
@pytest.mark.parametrize(["cyclonedx"], [("cdx16",), ("cdx15",)])
def test_translation(cyclonedx, spdx):
    tm = TranslationMap.from_file(
        "tests/testdata/test_translation/sample_spdx23_cdx16.yml"
    )
    first_doc = Document(
        get_mapping(f"tests/testdata/test_translation/sample_{spdx}.json")
    )
    second_doc = Document(
        get_mapping(f"tests/testdata/test_translation/sample_{cyclonedx}.json")
    )
    assert tm.convert(first_doc, SBOMFormat(cyclonedx)).doc == second_doc.doc
    assert tm.convert(second_doc, SBOMFormat(spdx)).doc == first_doc.doc


def test_spdx_product_translation(built_in_translation_map: TranslationMap):
    spdx_doc = Document.from_file(
        "tests/testdata/test_translation/actual_data/rhel-9.2-eus.spdx.json"
    )
    cdx_doc = built_in_translation_map.convert(spdx_doc)
    new_doc = built_in_translation_map.convert(cdx_doc)
    for doc in spdx_doc, new_doc:
        # Ignore these autogenerated fields:
        doc.doc.pop("documentNamespace")
        creators = doc.doc.get("creationInfo").get("creators")
        creators = [c for c in creators if not c.startswith("Tool: SBOMGrader")]
        doc.doc["creationInfo"]["creators"] = creators
    assert ordered(spdx_doc.doc) == ordered(new_doc.doc)


def test_cdx_product_translation(built_in_translation_map: TranslationMap):
    cdx_doc = Document.from_file(
        "tests/testdata/test_translation/actual_data/rhel-9.2-eus.cdx.json"
    )
    new_doc = built_in_translation_map.convert(
        built_in_translation_map.convert(cdx_doc)
    )
    for doc in cdx_doc, new_doc:
        # Ignore these autogenerated fields:
        doc.doc.pop("serialNumber")
        tools = doc.doc.get("metadata").get("tools").get("components", [])
        tools = [t for t in tools if t.get("name") != "SBOMGrader"]
        doc.doc["metadata"]["tools"] = tools
        prune(doc.doc)
    assert ordered(cdx_doc.doc) == ordered(new_doc.doc)


def test_spdx_rpm_translation(built_in_translation_map: TranslationMap):
    spdx_doc = Document.from_file(
        "tests/testdata/test_translation/actual_data/openssl-3.0.7-18.el9_2.spdx.json"
    )
    cdx_doc = built_in_translation_map.convert(spdx_doc)
    new_doc = built_in_translation_map.convert(cdx_doc)
    for doc in spdx_doc, new_doc:
        # Ignore these autogenerated fields:
        doc.doc.pop("documentNamespace")
        doc.doc.pop("files", None)
        creators = doc.doc.get("creationInfo").get("creators")
        creators = [c for c in creators if not c.startswith("Tool: SBOMGrader")]
        doc.doc["creationInfo"]["creators"] = creators
        for component in doc.doc.get("packages", []):
            for annotation in component.get("annotations", []):
                for key in "annotator", "annotationDate":
                    annotation.pop(key, None)
            component.pop("packageFileName", None)
        prune(doc.doc)
    assert ordered(spdx_doc.doc) == ordered(new_doc.doc)


def test_cdx_rpm_translation(built_in_translation_map: TranslationMap):
    cdx_doc = Document.from_file(
        "tests/testdata/test_translation/actual_data/openssl-3.0.7-18.el9_2.cdx.json"
    )
    spdx_doc = built_in_translation_map.convert(cdx_doc)
    new_doc = built_in_translation_map.convert(spdx_doc)
    for doc in cdx_doc, new_doc:
        # Ignore these autogenerated fields:
        doc.doc.pop("serialNumber")
        doc.doc.get("metadata", {}).get("component", {}).pop("pedigree", None)
        tools = doc.doc.get("metadata").get("tools").get("components", [])
        tools = [t for t in tools if t.get("name") != "SBOMGrader"]
        doc.doc["metadata"]["tools"] = tools
        prune(doc.doc)
    assert ordered(cdx_doc.doc) == ordered(new_doc.doc)


def test_cdx_image_translation(built_in_translation_map: TranslationMap):
    cdx_doc = Document.from_file(
        "tests/testdata/test_translation/actual_data/mandrel-for-jdk-21-rhel8-container-23.1-16_amd64.cdx.json"
    )
    spdx_doc = built_in_translation_map.convert(cdx_doc)
    new_doc = built_in_translation_map.convert(spdx_doc)
    for doc in cdx_doc, new_doc:
        # Ignore these autogenerated fields:
        doc.doc.pop("serialNumber")
        doc.doc.get("metadata").pop("supplier", None)
        to_remove = []
        for idx, dep in enumerate(doc.doc.get("dependencies")):
            if not dep.get("dependsOn"):
                to_remove.append(idx)
        for idx in reversed(to_remove):
            doc.doc["dependencies"].pop(idx)
        for component in doc.doc["components"]:
            component.get("supplier", {}).pop("url", None)
        doc.doc.get("metadata").get("component").get("supplier", {}).pop("url", None)
        tools = doc.doc.get("metadata").get("tools", {}).get("components", [])
        tools = [t for t in tools if t.get("name") != "SBOMGrader"]
        doc.doc["metadata"]["tools"] = tools
        prune(doc.doc)
    assert ordered(cdx_doc.doc) == ordered(new_doc.doc)



from sbomgrader.core.definitions import TRANSLATION_MAP_DIR
from sbomgrader.core.documents import Document
from sbomgrader.core.formats import SBOMFormat, get_fallbacks
from sbomgrader.core.utils import is_mapping
from sbomgrader.translate.translation_map import TranslationMap


def choose_map(
    document: Document, out: SBOMFormat, *custom_maps: TranslationMap
) -> TranslationMap:
    """Choose the translation map according to the document and output format."""
    for map_set in (custom_maps, get_default_maps()):
        # Prefer custom maps to default ones
        for match_type in ("exact", "suitable"):
            # Prefer exact matches to fallbacks
            for map_ in map_set:
                if getattr(map_, f"is_{match_type}_map")(document.sbom_format, out):
                    return map_
    raise NotImplementedError(
        f"Cannot convert from format {document.sbom_format.value} to {out.value}. No such map is implemented."
    )


def get_default_maps() -> list[TranslationMap]:
    """Returns the list of pre-installed maps."""
    ans = []
    for file in TRANSLATION_MAP_DIR.iterdir():
        if not is_mapping(file):
            continue
        ans.append(TranslationMap.from_file(file))
    return ans


def get_all_map_list_tuples(
    *custom_maps: TranslationMap,
) -> tuple[set[tuple[SBOMFormat, SBOMFormat]], set[tuple[SBOMFormat, SBOMFormat]]]:
    """
    Returns 2 tuples, each contains a pair of SBOMFormat instances.
    These represent available translation directions.
    """
    direction_tuples = set()
    all_maps = [*get_default_maps(), *custom_maps]
    for map_ in all_maps:
        formats = tuple(sorted((map_.first, map_.second), key=lambda x: x.value))
        direction_tuples.add(formats)
    # Add fallbacks, include transitive relations
    fallback_tuples = {*direction_tuples}
    starting_fallback_tuples = set()
    while starting_fallback_tuples != fallback_tuples:
        # Perform until no further changes occur
        starting_fallback_tuples = {*fallback_tuples}
        for tuple_ in starting_fallback_tuples:
            for format_ in SBOMFormat:
                fallbacks = get_fallbacks(format_)
                if tuple_[0] in fallbacks:
                    fallback_tuples.add((format_, tuple_[1]))
                if tuple_[1] in fallbacks:
                    fallback_tuples.add((tuple_[0], format_))
    fallback_tuples = fallback_tuples - direction_tuples
    return direction_tuples, fallback_tuples


def get_all_map_list_markdown(*custom_maps: TranslationMap) -> str:
    """Returns a Markdown string representing all possible translation directions."""
    map_directions, fallback_tuples = get_all_map_list_tuples(*custom_maps)
    ans = ""
    for tuple_ in map_directions:
        ans += f"- {tuple_[0].value} <-> {tuple_[1].value}\n"
    for tuple_ in fallback_tuples:
        ans += f"- {tuple_[0].value} <-> {tuple_[1].value} (fallback)\n"
    return ans



import re
import uuid
from typing import Any

import yaml
from packageurl import PackageURL
from spdx_tools.spdx.model.spdx_no_assertion import SPDX_NO_ASSERTION_STRING

from sbomgrader.core.definitions import FIELD_NOT_PRESENT
from sbomgrader.translation_maps.transformers.red_hat_spdx23_cdx16.utils import (
    SPDX_CDX_HASHES,
)

_CACHE = {}


def get_serial_number(_: Any) -> str:
    return str(uuid.uuid4())


def spdxid_to_bom_ref(spdxid: str, purls: list[str]) -> str:
    if spdxid in _CACHE:
        return _CACHE[spdxid]
    purl = next(iter(purls), FIELD_NOT_PRESENT)
    if purl is FIELD_NOT_PRESENT:
        purl = spdxid.replace("SPDXRef-", "", 1)
    _CACHE[spdxid] = purl
    return _CACHE[spdxid]


def hash_alg_from_spdx_to_cdx(algorithm: str) -> str:
    dic = {first: second for first, second in SPDX_CDX_HASHES}
    return dic.get(algorithm, algorithm)


def purl_to_component_type(purl: str) -> str | None | FIELD_NOT_PRESENT.__class__:
    if not isinstance(purl, str):
        return FIELD_NOT_PRESENT
    if purl.startswith("pkg:oci/"):
        return "container"
    if (
        purl.startswith("pkg:rpm/")
        or purl.startswith("pkg:maven/")
        or purl.startswith("pkg:generic/")
    ):
        return "library"
    return FIELD_NOT_PRESENT


def namespaces_to_url(namespaces: list[str]) -> list[str]:
    ans = []
    for namespace in namespaces:
        match = re.match(r"https?://[\w.-]+", namespace)
        if match:
            ans.append(match.group())

    return ans


def sanitise_supplier(supplier: str) -> str:
    if not isinstance(supplier, str):
        return FIELD_NOT_PRESENT
    return re.sub(r"^\w+:\s", "", supplier, count=1)


def tool_to_dict(tool_string: str) -> dict[str, str]:
    if not isinstance(tool_string, str):
        return FIELD_NOT_PRESENT
    tool_string = tool_string.strip()
    tool_string = re.sub(r"Tool: ", "", tool_string, flags=re.IGNORECASE, count=1)

    splits = tool_string.rsplit(" ", 1)
    if len(splits) == 2 and re.search(r"\d", splits[1]):
        # Let's assume that the version is the last bit and contains numbers
        return {"name": splits[0], "version": splits[1]}
    else:
        return {"name": tool_string}


def annotations_to_properties(annotations: list[str]) -> list[dict[str, str]]:
    ans = []
    for annotation in annotations:
        try:
            dic = yaml.safe_load(annotation)
            tmp = {}
            for key, value in dic.items():
                tmp.update({"name": key, "value": value})
            ans.append(tmp)
        except Exception:
            ans.append({"name": "COMMENT", "value": annotation})
    return ans


def purl_with_download_location(purl: str, download_location_var: list[str]) -> str:
    if not purl or not isinstance(purl, str):
        return FIELD_NOT_PRESENT
    purl_obj = PackageURL.from_string(purl)
    if "download_url" in purl_obj.qualifiers:
        return purl
    download_location = next(iter(download_location_var), SPDX_NO_ASSERTION_STRING)
    if download_location == SPDX_NO_ASSERTION_STRING:
        return purl
    purl_obj.qualifiers["download_url"] = download_location
    return purl_obj.to_string()



import json

from cyclonedx.schema import SchemaVersion
from cyclonedx.validation.json import JsonValidator

from sbomgrader.core.definitions import FIELD_NOT_PRESENT


def validate_schema(doc: dict):
    validator = JsonValidator(SchemaVersion.V1_5)
    error = validator.validate_str(json.dumps(doc))
    if error:
        raise AssertionError(error.data)


def package_relationships(doc: dict):
    all_bom_refs_from_deps = set()
    for dep in doc.get("dependencies", []):
        all_bom_refs_from_deps.update(dep.get("dependsOn", []))
        all_bom_refs_from_deps.add(dep["ref"])
    for idx, component in enumerate(doc.get("components", [])):
        assert (
            component.get("bom-ref", FIELD_NOT_PRESENT) in all_bom_refs_from_deps
        ), f"Test failed item components[{idx}]: Component does not have any relationships to other components: {component}"



import base64
import re
import urllib.parse
import uuid

from packageurl import PackageURL
from spdx_tools.spdx.model.spdx_no_assertion import SPDX_NO_ASSERTION_STRING

from sbomgrader.core.definitions import FIELD_NOT_PRESENT
from sbomgrader.translation_maps.transformers.red_hat_spdx23_cdx16.utils import (
    SPDX_CDX_HASHES,
)

_CACHE = {}


def bom_ref_to_spdxid(bom_ref: str, component_name_var: list[str]) -> str:

    if bom_ref in _CACHE:
        return _CACHE[bom_ref]
    if not bom_ref or bom_ref is FIELD_NOT_PRESENT:
        name = next(iter(component_name_var), None)
        if name and name is not FIELD_NOT_PRESENT:
            bom_ref = name
        else:
            bom_ref = base64.b64encode(uuid.uuid4().bytes)

    _CACHE[bom_ref] = f"SPDXRef-{re.sub(r"[^A-Za-z\d.-]", "-", bom_ref)}"
    return _CACHE[bom_ref]


def hash_alg_from_cdx_to_spdx(algorithm: str) -> str:
    dic = {second: first for first, second in SPDX_CDX_HASHES}
    return dic.get(algorithm, algorithm)


def url_to_namespace(
    url: str,
    component_name_var: list[str],
    serial_no_var: list[str],
) -> str:
    if not url or not isinstance(url, str):
        url = "https://github.com/BorekZnovustvoritel/SBOM-Grader"
    end_string = urllib.parse.quote_plus(
        next(iter(component_name_var), FIELD_NOT_PRESENT)
        + "/"
        + next(iter(serial_no_var), FIELD_NOT_PRESENT)
    )

    return url + "/" + end_string


def cpe_to_cpe_type(cpe: str) -> str:
    if cpe.startswith("cpe:2.3:"):
        return "cpe23Type"
    return "cpe22Type"


def purl_to_download_location(purl: str) -> str:
    if not purl or not isinstance(purl, str):
        return SPDX_NO_ASSERTION_STRING
    purl_obj = PackageURL.from_string(purl)
    loc = purl_obj.qualifiers.get("download_url") or SPDX_NO_ASSERTION_STRING
    return loc



import json
import sys
from copy import copy
from dataclasses import fields, dataclass, field
from pathlib import Path
from typing import Iterable, Any

import jsonschema.exceptions
import yaml

from sbomgrader.grade.cookbooks import Cookbook, CookbookResult
from sbomgrader.core.definitions import COOKBOOKS_DIR
from sbomgrader.core.documents import Document
from sbomgrader.core.enums import SBOMType, SBOMTime, OutputType, Grade
from sbomgrader.grade.rules import RuleSet, Result


@dataclass
class CookbookBundleResult:
    cookbook_bundle: "CookbookBundle"
    cookbook_results: list[CookbookResult] = field(default_factory=list)

    def output(self, o_type: OutputType) -> str:
        if o_type in {OutputType.MARKDOWN, OutputType.VISUAL}:
            ans = "# Cookbook bundle result\n\n"
            ans += f"**Grade: {self.grade.value}**\n\n"
            ans += "## Used cookbooks\n\n"
            for cookbook_result in self.cookbook_results:
                ans += f"- {cookbook_result.cookbook.name}\n"
            ans += "\n---\n\n"
            ans += "\n---\n\n".join(
                cookbook_result.output(o_type)
                for cookbook_result in self.cookbook_results
            )
            return ans
        if o_type is OutputType.JSON:
            return json.dumps(self.to_dict(), indent=4)
        return yaml.dump(self.to_dict())

    @property
    def grade(self) -> Grade:
        if decisive_cookbook := self.cookbook_bundle.decisive_cookbook:
            cookbook_result: CookbookResult = next(
                filter(
                    lambda x: x.cookbook.name == decisive_cookbook,
                    self.cookbook_results,
                ),
                None,
            )
            if cookbook_result is not None:
                return cookbook_result.grade
        grades = [x.grade for x in self.cookbook_results]
        return sorted(grades, key=lambda x: ord(x.value))[-1]

    def to_dict(self) -> dict[str, Any]:
        ans = {"cookbook_results": [], "grade": self.grade.value}
        for cookbook_result in self.cookbook_results:
            ans["cookbook_results"].append(cookbook_result.to_dict())
        return ans

    def __iter__(self):
        yield from self.cookbook_results


class CookbookBundle:
    def __init__(
        self, cookbooks: Iterable[Cookbook], decisive_cookbook: str | None = None
    ):
        self.cookbooks = set(cookbooks)
        self.decisive_cookbook: str | None = decisive_cookbook

    @staticmethod
    def from_directory(dir_path: Path) -> "CookbookBundle":
        cookbook_bundle = CookbookBundle([])
        if dir_path.is_dir():
            for entity in dir_path.iterdir():
                if not entity.is_file() or (
                    not entity.name.endswith(".yml")
                    and not entity.name.endswith(".yaml")
                ):
                    continue
                try:
                    cookbook_bundle += Cookbook.from_file(entity)
                except jsonschema.exceptions.ValidationError:
                    print(f"Could not load file {entity.absolute()}", file=sys.stderr)
        return cookbook_bundle

    @property
    def all_rules(self) -> set[str]:
        all_rules: set[str] = set()
        for cookbook in self.cookbooks:
            all_rules.update(cookbook.must)
            all_rules.update(cookbook.should)
            all_rules.update(cookbook.may)
        return all_rules

    @property
    def ruleset(self) -> RuleSet:
        ruleset = RuleSet()
        for cookbook in self.cookbooks:
            ruleset += cookbook.ruleset
        return ruleset

    def __call__(self, doc: Document) -> CookbookBundleResult:
        result = self.ruleset(doc)
        ans = []
        for cookbook in self.cookbooks:
            kwargs = {}
            for attr_obj in fields(Result):
                attr = attr_obj.name
                attr_value = getattr(result, attr)
                if isinstance(attr_value, dict):
                    kwargs[attr] = {
                        k: v for k, v in attr_value.items() if k in cookbook
                    }
                else:
                    kwargs[attr] = {v for v in attr_value if v in cookbook}
            new_result = Result(**kwargs)
            ans.append(CookbookResult(new_result, cookbook))
        return CookbookBundleResult(self, ans)

    @staticmethod
    def for_document_type(
        sbom_type: SBOMType, requested_stage: SBOMTime = SBOMTime.UNSPECIFIED
    ) -> "CookbookBundle":
        cookbook_identifiers = []
        if sbom_type is SBOMType.PRODUCT:
            cookbook_identifiers.append(COOKBOOKS_DIR / (sbom_type.value + ".yml"))
            decisive_cookbook = sbom_type.value

        elif requested_stage is SBOMTime.UNSPECIFIED:
            for sbom_time in SBOMTime.RELEASE, SBOMTime.BUILD:
                cookbook_identifiers.append(
                    COOKBOOKS_DIR / f"{sbom_type.value}_{sbom_time.value}.yml"
                )
            decisive_cookbook = f"{sbom_type.value}_{SBOMTime.RELEASE}"
        else:
            cookbook_identifiers.append(
                COOKBOOKS_DIR / f"{sbom_type.value}_{requested_stage.value}.yml"
            )
            decisive_cookbook = f"{sbom_type.value}_{requested_stage.value}"
        return CookbookBundle(
            [Cookbook.from_file(identifier) for identifier in cookbook_identifiers],
            decisive_cookbook,
        )

    def __add__(self, other):
        if isinstance(other, Cookbook):
            new_bundle = CookbookBundle(copy(self.cookbooks))
            new_bundle.cookbooks.add(other)
            return new_bundle
        if isinstance(other, CookbookBundle):
            new_bundle = CookbookBundle(copy(self.cookbooks))
            new_bundle.cookbooks.update(other.cookbooks)
            return new_bundle
        raise TypeError(f"Cannot add types Cookbook and {type(other)}.")

    def __iter__(self) -> Cookbook:
        yield from self.cookbooks



from pathlib import Path

from sbomgrader.core.cached_python_loader import PythonLoader


class RuleLoader(PythonLoader):
    def __init__(self, implementation: str, *file_references: str | Path):
        super().__init__(*file_references)
        self.implementation: str = implementation



import re
import sys
from collections import defaultdict
from dataclasses import dataclass
from functools import cached_property
from typing import Union, Any, Callable


from sbomgrader.core.definitions import (
    FIELD_NOT_PRESENT,
    FieldNotPresentError,
    MAX_ITEM_PREVIEW_LENGTH,
    START_PREVIEW_CHARS,
    END_PREVIEW_CHARS,
    VAR_REF_REGEX,
)
from sbomgrader.core.enums import QueryType


class PathParser:
    """Parses the FieldPath expression into an iterable list."""

    def __init__(self, path: str):
        self._path = path
        self.__next_is_query = False
        self.ans: dict[str, list[Union[str, QueryParser]]] = defaultdict(list)

    def __create_field(
        self, field: str, next_is_query: bool, relative_path: "PathParser"
    ) -> None:
        if self.__next_is_query:
            if field == "@":
                try:
                    appropriate_relative_field = relative_path.parse()[
                        len(self.ans[relative_path.raw_path])
                    ]
                except IndexError:
                    raise ValueError(
                        f"Problem parsing path '{self._path}' with relative hint '{relative_path.raw_path}'. "
                        f"The relative hint is too short!"
                    )
                if isinstance(appropriate_relative_field, QueryParser):
                    index_query = next(iter(appropriate_relative_field.parse()), None)
                    if not index_query:
                        raise ValueError(
                            f"Problem parsing path '{self._path}' with relative hint '{relative_path.raw_path}'. "
                            f"There are no queries to follow at step '{field}'."
                        )
                    if index_query.type_ is not QueryType.INDEX:
                        raise ValueError(
                            f"Problem parsing path '{self._path}' with relative hint '{relative_path.raw_path}'. "
                            f"This is not an index query type: '{index_query.type_}'"
                        )
                    field = index_query.value
            next_ = QueryParser(field)
        else:
            next_ = field.strip()
        self.__next_is_query = next_is_query

        if next_:
            self.ans[relative_path.raw_path].append(next_)

    def parse(
        self, relative_path: str | None = None
    ) -> list[Union[str, "QueryParser"]]:
        """
        Returns a list of dictionary field names and queries for list filtering.
        :argument relative_path: Used to populate "relative" fields
        (relative path has the '@' symbol).
        """
        relative_path = relative_path or ""
        if relative_path in self.ans:
            return self.ans[relative_path]
        parsed_relative_path = PathParser(relative_path)
        if relative_path:
            if any(
                any(query.type_ != QueryType.INDEX for query in item.parse())
                for item in parsed_relative_path.parse()
                if isinstance(item, QueryParser)
            ):
                raise ValueError(
                    "Relative path can only include field names and absolute indices. "
                    "HINT: If you wish to get a list of absolute paths, use the method "
                    "`FieldResolver.get_paths()`."
                )
        resolve_path = self._path
        if resolve_path.startswith("@"):
            if not relative_path:
                raise ValueError(
                    "Cannot resolve relative path if no relative path is passed!"
                )
            resolve_path = resolve_path.replace("@", relative_path, 1)
        in_block = 1 if self.__next_is_query else 0
        buffer = ""
        for char in resolve_path:
            if char == "[":
                if not in_block:
                    self.__create_field(buffer, True, parsed_relative_path)
                    buffer = ""
                else:
                    buffer += char
                in_block += 1
            elif char == "]":
                in_block -= 1
                if not in_block:
                    self.__create_field(buffer, False, parsed_relative_path)
                    buffer = ""
                else:
                    buffer += char
            elif char == ".":
                if not in_block:
                    self.__create_field(buffer, False, parsed_relative_path)
                    buffer = ""
                else:
                    buffer += char
            else:
                buffer += char
        if buffer:
            self.__create_field(buffer, False, parsed_relative_path)
        self.__next_is_query = None
        return self.ans[relative_path]

    def __eq__(self, other):
        if not isinstance(other, PathParser):
            raise TypeError(
                f"Cannot compare PathParser to object of type {type(other)}"
            )
        return self.parse() == other.parse()

    @property
    def raw_path(self) -> str:
        """The original string containing the path."""
        return self._path

    @property
    def variable_references(self) -> set[str]:
        parsed_path = self.parse()
        ans = set()
        for step in parsed_path:
            if isinstance(step, QueryParser):
                ans.update(step.variable_references)
        return ans


@dataclass
class Query:
    """Class holding an information about list filtering."""

    type_: QueryType
    value: str | int | None
    field_path: PathParser | None

    @property
    def variable(self) -> str | None:
        if (
            self.value
            and isinstance(self.value, str)
            and (match := re.match(r"^\$\{(?P<varname>\w+)}$", self.value))
        ):
            return match.group("varname")


class QueryParser:
    """
    Parses strings into a list of Queries.
    """

    def __init__(self, path: str | int):
        self._path = path
        self.ans: dict[str, list[Query]] = defaultdict(list)

    def __eq__(self, other):
        if not isinstance(other, QueryParser):
            raise TypeError(
                f"Cannot compare QueryParser to object of type {type(other)}"
            )
        return self.parse() == other.parse()

    def parse(self, relative_path_index: str | None = None) -> list[Query]:
        """Parse the query list. If required, replaces the relative symbol '@' with the provided index."""
        if relative_path_index in self.ans:
            return self.ans[relative_path_index]
        if isinstance(self._path, int):
            return [Query(QueryType.INDEX, value=self._path, field_path=None)]
        queries = []
        field_buffer = ""
        operation_buffer = ""
        value_buffer = ""
        in_block = 0
        in_operation = False
        after_operation = False
        operation_symbols = {"!", "=", "%", "|", "&"}
        for char in self._path:
            if re.match(r"\s", char) and not after_operation:
                continue
            if char not in operation_symbols and in_operation:
                after_operation = True
            if char in operation_symbols and not in_block and not after_operation:
                operation_buffer += char
                in_operation = True

            elif after_operation and char != ",":
                value_buffer += char
            elif char == "," and in_operation:
                queries.append(
                    Query(
                        type_=QueryType(operation_buffer),
                        field_path=(
                            None if not field_buffer else PathParser(field_buffer)
                        ),
                        value=None if not value_buffer else value_buffer,
                    )
                )
                field_buffer = ""
                operation_buffer = ""
                value_buffer = ""
                in_operation = False
                after_operation = False
            elif char == "[":
                field_buffer += char
                in_block += 1
            elif char == "]":
                in_block -= 1
                field_buffer += char

            else:
                field_buffer += char.strip()

        if field_buffer or operation_buffer or value_buffer:
            if field_buffer == "@" and not operation_buffer and not value_buffer:
                # There is no query, just a relative symbol
                query = Query(
                    type_=QueryType.INDEX,
                    field_path=None,
                    value=int(relative_path_index),
                )
            elif (
                (m := re.fullmatch(r"\d+", field_buffer))
                and not operation_buffer
                and not value_buffer
            ):
                # There is no query, just an index
                query = Query(
                    type_=QueryType.INDEX, field_path=None, value=int(m.group())
                )
            else:
                query = Query(
                    type_=QueryType(operation_buffer.strip()),
                    field_path=PathParser(field_buffer.strip()),
                    value=self._load_val(value_buffer),
                )
            queries.append(query)
        self.ans[relative_path_index] = queries
        return queries

    @property
    def variable_references(self) -> set[str]:
        parsed_queries = self.parse()
        ans = set()
        for query in parsed_queries:
            if var_name := query.variable:
                ans.add(var_name)
            if sub_path_parser := query.field_path:
                ans.update(sub_path_parser.variable_references)
        return ans

    @staticmethod
    def _load_val(value: str) -> Any:
        stripped_val = value.strip()
        if stripped_val == FIELD_NOT_PRESENT.string_repr:
            return FIELD_NOT_PRESENT
        return stripped_val

    def __repr__(self):
        return str(self._path)


class Variable:
    def __init__(
        self,
        name: str,
        field_path: str,
    ):
        self.name = name
        self.raw_field_path = field_path
        self.path_parser = PathParser(self.raw_field_path)

    @staticmethod
    def from_schema(schema_list: list[dict[str, Any]]) -> dict[str, "Variable"]:
        """Load a variable from a dictionary."""
        ans = {}
        if not schema_list:
            return ans
        for item in schema_list:
            name = item["name"]
            field_path = item["fieldPath"]
            ans[name] = Variable(name, field_path)
        return ans

    @property
    def is_relative(self) -> bool:
        """Does the definition of this Variable contain a relative path?"""
        return self.is_fully_relative or self.is_partially_relative

    @property
    def is_fully_relative(self):
        return self.raw_field_path.startswith("@")

    @property
    def is_partially_relative(self):
        return "[@]" in self.raw_field_path

    @property
    def without_relative_start(self) -> "Variable":
        if not self.is_fully_relative:
            return self
        return Variable(self.name, self.path_parser.raw_path.removeprefix("@."))

    @cached_property
    def dependencies(self) -> set[str]:
        deps = {
            match.group("var_id")
            for match in re.finditer(VAR_REF_REGEX, self.raw_field_path)
        }
        assert self.name not in deps, f"Self referencing variable {self.name} found."
        return deps

    def __hash__(self):
        return self.name.__hash__()

    def __repr__(self):
        return f"<{self.__class__.__name__}, name: {self.name}, field_path: {self.raw_field_path}>"


class FieldResolver:
    """
    Resolves path expressions and dictionaries (documents).
    Can return their paths, values or executes functions on each occurrence.
    """

    def __init__(self, variables: dict[str, Variable]):
        self._uninitialized_vars = variables

    @property
    def var_definitions(self) -> dict[str, Variable]:
        return self._uninitialized_vars

    @property
    def absolute_variables(self) -> dict[str, Variable]:
        """Returns just variables without a relative path definition."""
        return {
            key: val
            for key, val in self._uninitialized_vars.items()
            if not val.is_relative
        }

    @property
    def relative_variables(self) -> dict[str, Variable]:
        """Returns just variables with a relative path definition."""
        return {
            key: val for key, val in self._uninitialized_vars.items() if val.is_relative
        }

    @property
    def fully_relative_variables(self) -> dict[str, Variable]:
        """
        Returns just variables with a fully relative path definition.
        (Their path starts with a '@', they search through
         direct descendants of the relative path).
        """
        return {
            key: val for key, val in self._uninitialized_vars.items() if val.is_relative
        }

    def __find_dependencies_for_subset(
        self,
        subset_of_variables: list[str],
        already_resolved: set[str] | dict[str, Any],
    ) -> set[str]:
        """Returns dependencies for a subset of variables."""
        to_resolve = set(subset_of_variables)
        ans = set(already_resolved)
        while to_resolve:
            var_name = to_resolve.pop()
            variable = self.var_definitions.get(var_name)
            if not variable:
                continue
            for dep_name in variable.dependencies:
                if dep_name not in ans:
                    to_resolve.add(dep_name)
            ans.add(var_name)
        return ans

    @staticmethod
    def __mark_variable_as_resolved(
        dependency_dict: dict[str, set[str]], var_name: str
    ):
        dependency_dict.pop(var_name, None)
        for deps in dependency_dict.values():
            if var_name in deps:
                deps.remove(var_name)

    def resolve_variables(
        self,
        whole_doc: dict[str, Any],
        path_to_instance: str | None = None,
        already_resolved_variables: dict[str, list[Any]] = None,
        warning_on: bool = True,
        variables_needed: list[str] | set[str] = None,
        path_prefix: str = "",
    ) -> dict[str, list[Any]]:
        """
        Resolve dependencies.
        Without the argument `path_to_instance` this method cannot resolve relative variables
        nor absolute variables relying on relative ones.
        :argument whole_doc: The document that the FieldPath expression should be evaluated on
        :argument path_to_instance: The base for the relative paths.
        :argument already_resolved_variables: Variable values resolved previously. Helps with
        performance.
        :argument warning_on: Should this function display a warning to the STDERR?
        :argument variables_needed: Specify a subset of variables that need resolving.
        :argument path_prefix: Optionally provide a path that will be prepended to each "path tried".
        """
        already_resolved_variables = already_resolved_variables or {}
        if not variables_needed:
            # first resolve dependency tree for variables
            if path_to_instance:
                vars_to_resolve = self._uninitialized_vars
            else:
                vars_to_resolve = self.absolute_variables
            vars_to_resolve = {
                k: v
                for k, v in vars_to_resolve.items()
                if k not in already_resolved_variables
            }
        else:
            vars_to_resolve = {
                k: self.var_definitions[k]
                for k in self.__find_dependencies_for_subset(
                    variables_needed, already_resolved_variables
                )
            }
        # This will keep track of variable values that need to be
        # included for other variable resolution
        all_dependencies: dict[str, set[str]] = {
            var_name: variable_def.dependencies
            for var_name, variable_def in vars_to_resolve.items()
        }
        # This variable will be reduced so only the outstanding
        # (blocking, unresolved) dependencies are included
        dependencies: dict[str, set[str]] = {
            var_name: set(deps) for var_name, deps in all_dependencies.items()
        }
        # This variable keeps track of what is already resolved
        resolved_variables: dict[str, list] = {**already_resolved_variables}
        for var_name in resolved_variables:
            self.__mark_variable_as_resolved(dependencies, var_name)
        while not all(var_name in resolved_variables for var_name in dependencies):
            # Get a var with no dependencies
            var_name, var_deps = sorted(dependencies.items(), key=lambda x: len(x[1]))[
                0
            ]
            if not path_to_instance and any(
                self._uninitialized_vars[dep_name].is_relative for dep_name in var_deps
            ):
                # Cannot resolve absolute variable referencing a relative one
                self.__mark_variable_as_resolved(dependencies, var_name)
                continue
            assert not var_deps, (
                f"Circular variable reference found for variable {var_name}. "
                f"Needs to resolve: {dependencies[var_name]}. "
                f"Already resolved: {set(resolved_variables.keys())}"
            )

            resolved_variables[var_name] = []

            def add_to_variable(value: Any, _) -> None:
                resolved_variables[var_name].append(value)

            path = vars_to_resolve[var_name].path_parser.parse(path_to_instance)
            variables_needed = self.__cast_vars_to_sets(
                {
                    dep_name: dep_value
                    for dep_name, dep_value in resolved_variables.items()
                    if dep_name in all_dependencies.get(var_name, set())
                }
            )
            try:
                self._run_on_path(
                    whole_doc,
                    path,
                    variables_needed,
                    path_prefix,
                    add_to_variable,
                    False,
                )
            except Exception as e:
                if warning_on:
                    print(
                        f"Could not parse variable {var_name}, problem: {str(e)}",
                        file=sys.stderr,
                    )

            self.__mark_variable_as_resolved(dependencies, var_name)
        return resolved_variables

    @staticmethod
    def __add_at_path(
        mutable_doc: dict[str, Any],
        path_remaining: list[str | QueryParser | PathParser],
    ):
        if path_remaining:
            step = path_remaining[0]
        else:
            return
        if path_remaining[1:] and isinstance(path_remaining[1], str):
            # Add a dict
            mutable_doc[step] = {}
        if path_remaining[1:] and isinstance(path_remaining[1], QueryParser):
            # Add a list
            mutable_doc[step] = []

    @staticmethod
    def __cast_vars_to_sets(
        variables: dict[str, list[Any]],
    ) -> dict[str, list[Any] | set[Any]]:
        """
        Tries to convert lists of values to
        sets of values. Leaves the original value
        if it contains unhashable objects.
        """
        new_variables = {}
        for var_name in variables:
            try:
                new_value = set(variables[var_name])
                new_variables[var_name] = new_value
            except TypeError:
                new_variables[var_name] = variables[var_name]
        return new_variables

    def _run_on_path(
        self,
        doc_: Union[dict, list[Any], FIELD_NOT_PRESENT],
        path: list[str | QueryParser | PathParser],
        variable_values: dict[str, list[Any] | set[Any]],
        path_tried: str,
        func_to_run: Callable[[Any, str], Any],
        accept_not_present_field: bool,
        create_nonexistent: bool = False,
    ):
        """
        This function is the main resolver. Recursively calls itself on nested fields
        to search for all occurrences.
        """
        if not accept_not_present_field and doc_ is FIELD_NOT_PRESENT:
            raise FieldNotPresentError("Field not present: ", path_tried)
        if not path or doc_ is FIELD_NOT_PRESENT:
            # The path has ended
            try:
                resp = func_to_run(doc_, path_tried)
                assert resp is True or resp is None
            except Exception as e:
                item_str = str(doc_)
                if len(item_str) > MAX_ITEM_PREVIEW_LENGTH:
                    item_str = f"{item_str[:START_PREVIEW_CHARS]}...{item_str[-END_PREVIEW_CHARS:]}"
                if not path_tried:
                    path_tried = "."
                message_to_return = (
                    f"Check did not pass for item: {item_str} at path: {path_tried}\n"
                    + "\n".join(str(m) for m in e.args)
                )
                raise type(e)(message_to_return) from e
            return
        step = path[0]
        if isinstance(step, str):
            # Field name
            assert isinstance(
                doc_, dict
            ), f"Cannot access field '{step}' on other objects than dicts. Provided object: {doc_}"
            if step == "?":
                assert path[1:] and isinstance(
                    path[1], str
                ), "Cannot use ? before anything else than a field name."
                if path[1] not in doc_ and create_nonexistent:
                    self.__add_at_path(doc_, path[1:])
                if path[1] in doc_:
                    self._run_on_path(
                        doc_,
                        path[1:],
                        variable_values,
                        path_tried,
                        func_to_run,
                        accept_not_present_field,
                        create_nonexistent,
                    )
            else:
                if create_nonexistent and step not in doc_:
                    self.__add_at_path(doc_, path)
                self._run_on_path(
                    doc_.get(step, FIELD_NOT_PRESENT),
                    path[1:],
                    variable_values,
                    path_tried + f".{step}",
                    func_to_run,
                    accept_not_present_field,
                    create_nonexistent,
                )
        elif isinstance(step, QueryParser):
            assert isinstance(
                doc_, list
            ), f"Queries can only be performed on lists! Tested path: {path_tried}, item: {doc_}"
            queries = step.parse()

            to_use = []
            can_fail_for_some = False

            for query in queries:
                if query.type_ in {QueryType.EACH, QueryType.ANY}:
                    # Use every list index available
                    to_use.append(set(range(len(doc_))))
                    if query.type_ is QueryType.ANY:
                        can_fail_for_some = True
                    continue
                elif query.type_ is QueryType.INDEX:
                    to_use.append({query.value})
                    continue
                # Actually filter the list
                to_use_in_query = set()
                for idx, item in enumerate(doc_):
                    varname = query.variable
                    if query.type_ is QueryType.EQ:
                        if varname:
                            func = lambda x: x in variable_values[varname]
                        else:
                            func = lambda x: x == query.value
                    elif query.type_ is QueryType.NEQ:
                        if varname:
                            func = lambda x: x not in variable_values[varname]
                        else:
                            func = lambda x: x != query.value
                    elif query.type_ is QueryType.STARTSWITH:
                        if varname:
                            func = lambda x: isinstance(x, str) and any(
                                x.startswith(val) for val in variable_values[varname]
                            )
                        else:
                            func = lambda x: isinstance(x, str) and x.startswith(
                                query.value
                            )
                    elif query.type_ is QueryType.ENDSWITH:
                        if varname:
                            func = lambda x: isinstance(x, str) and any(
                                x.endswith(val) for val in variable_values[varname]
                            )
                        else:
                            func = lambda x: isinstance(x, str) and x.endswith(
                                query.value
                            )
                    elif query.type_ is QueryType.CONTAINS:
                        if varname:
                            func = lambda x: isinstance(x, str) and any(
                                val in x for val in variable_values[varname]
                            )
                        else:
                            func = lambda x: isinstance(x, str) and query.value in x
                    elif query.type_ is QueryType.NOT_CONTAINS:
                        if varname:
                            func = lambda x: isinstance(x, str) and all(
                                val not in x for val in variable_values[varname]
                            )
                        else:
                            func = lambda x: isinstance(x, str) and query.value not in x
                    final_func = lambda x, _: (
                        to_use_in_query.add(idx) if func(x) else None
                    )
                    parsed_path = (
                        query.field_path.parse() if query.field_path is not None else []
                    )
                    self._run_on_path(
                        item,
                        parsed_path,
                        variable_values,
                        path_tried + f"[{idx}]",
                        final_func,
                        True,
                        create_nonexistent,
                    )
                    to_use.append(to_use_in_query)
            to_use_final = set.intersection(*to_use) if to_use else {}
            failed = 0
            assertions = []
            for idx, item in enumerate(doc_):
                if idx not in to_use_final:
                    continue

                if can_fail_for_some:
                    try:
                        self._run_on_path(
                            item,
                            path[1:],
                            variable_values,
                            path_tried + f"[{idx}]",
                            func_to_run,
                            accept_not_present_field,
                            create_nonexistent,
                        )
                    except (AssertionError, FieldNotPresentError) as e:
                        failed += 1
                        assertions.append(e)
                    assert failed < len(
                        to_use_final
                    ), f"Check did not pass for any fields. Assertions: {assertions}, path: {path_tried}"
                else:
                    self._run_on_path(
                        item,
                        path[1:],
                        variable_values,
                        path_tried + f"[{idx}]",
                        func_to_run,
                        accept_not_present_field,
                        create_nonexistent,
                    )

    @staticmethod
    def ensure_field_path(
        field_path: str | list[Union[str, QueryParser]],
    ) -> list[Union[str, QueryParser]]:
        """Makes sure the FieldPath is in the parsed format."""
        return (
            field_path
            if isinstance(field_path, list)
            else PathParser(field_path).parse()
        )

    @staticmethod
    def __get_vars_from_path(
        path: str | list[str | QueryParser] | PathParser,
    ) -> set[str]:
        if isinstance(path, str):
            path = PathParser(path)
        if isinstance(path, PathParser):
            path = path.parse()
        if not isinstance(path, list):
            raise TypeError(f"Invalid path type supplied: {type(path)}")
        ans = set()
        for step in path:
            if isinstance(step, QueryParser):
                ans.update(step.variable_references)
        return ans

    def __populate_variables(
        self,
        doc: dict[str, Any],
        fallback_values: dict[str, Any],
        field_path: list[str | QueryParser] | str | PathParser,
        allow_fail: bool = False,
        prefer_fallback: bool = False,
    ) -> dict[str, Any]:
        variables_needed = self.__get_vars_from_path(field_path)
        args = {
            "whole_doc": doc,
            "warning_on": not allow_fail,
            "variables_needed": variables_needed,
        }
        if prefer_fallback:
            args["already_resolved_variables"] = fallback_values
        variables = {} if not fallback_values else {**fallback_values}
        variables.update(self.resolve_variables(**args))
        return variables

    def run_func(
        self,
        doc: dict[str, Any],
        func: Callable[[Any], Any],
        field_path: str | list[Union[str, QueryParser]],
        minimal_runs: int = 1,
        fallback_variables: dict[str, Any] | None = None,
        create_nonexistent: bool = False,
        path_prefix: str = "",
    ) -> None:
        """
        Execute a function on each field matching the FieldPath expression.
        :argument doc: The dictionary that the expression is evaluated on.
        :argument func: The function to run on each occurrence.
        :argument field_path: The FieldPath expression to locate all fields.
        :argument minimal_runs: If the number of executions of the function
        is not met, this method raises an Assertion Error.
        :argument fallback_variables: Variable values resolved previously in
        the format {"var_name": [var_value1, var_value2,...]}
        :argument create_nonexistent: If the path does not exist yet, should
        this function create it? Useful for document creation.
        :argument path_prefix: Optionally provide a path that will be prepended to each "path tried".
        """
        ran_on = set()

        def adjusted_func(value: Any, path: str) -> None:
            ran_on.add(path)
            func(value)

        parsed_path = self.ensure_field_path(field_path)
        resolved_variables = self.__cast_vars_to_sets(
            self.__populate_variables(
                doc, fallback_variables, parsed_path, create_nonexistent
            )
        )
        self._run_on_path(
            doc,
            parsed_path,
            resolved_variables,
            path_prefix,
            adjusted_func,
            create_nonexistent,
            create_nonexistent,
        )
        assert (
            len(ran_on) >= minimal_runs
        ), "Test was not performed on any fields because no fields match given filters."

    def get_paths(
        self,
        doc: dict[str, Any],
        field_path: str | list[Union[str, QueryParser]],
        fallback_variables: dict[str, Any],
        prefer_fallback: bool = False,
        path_prefix: str = "",
    ) -> list[str]:
        """
        Get paths to occurrences of fields matching the general expressions.
        Returns a list of concrete expressions (all indexes in a list are returned
        separately).
        :argument doc: The document to evaluate this expression on.
        :argument field_path: The FieldPath expression.
        :argument fallback_variables: The already-resolved variable values in
        the format {"var_name": [var_value1, var_value2,...]}
        :argument prefer_fallback: If set to `True`, fallback variables will not be overridden.
        This improves performance. Default is `False`.
        :argument path_prefix: Optionally provide a path that will be prepended to each "path tried".
        :return: a list of concrete occurrences (paths in the string format).
        """
        parsed_path = self.ensure_field_path(field_path)
        resolved_variables = self.__cast_vars_to_sets(
            self.__populate_variables(
                doc, fallback_variables, parsed_path, prefer_fallback
            )
        )
        try:

            paths = set()
            self._run_on_path(
                doc,
                parsed_path,
                resolved_variables,
                path_prefix,
                lambda _, path: paths.add(path),
                False,
                False,
            )
            return list(paths)
        except FieldNotPresentError:
            return []

    def get_objects(
        self,
        doc: dict[str, Any],
        field_path: str | list[Union[str, QueryParser]],
        fallback_variables: dict[str, Any] | None = None,
        create_nonexistent: bool = False,
        path_prefix: str = "",
    ) -> list[Any]:
        """
        Gets all fields matching the FieldPath expression.
        :argument doc: The document the expression is evaluated on.
        :argument field_path: The FieldPath expression.
        :argument fallback_variables: Already resolved variable values in
        the format {"var_name": [var_value1, var_value2,...]}.
        :argument create_nonexistent: If the fields do not exist already,
        should they be created? Useful for document creation.
        :argument path_prefix: Optionally provide a path that will be prepended to each "path tried".
        :return: A list of field values.
        """
        ans = []

        def add_to_ans(item: Any) -> None:
            ans.append(item)

        try:
            self.run_func(
                doc,
                add_to_ans,
                field_path,
                minimal_runs=0,
                fallback_variables=fallback_variables,
                create_nonexistent=create_nonexistent,
                path_prefix=path_prefix,
            )
            return ans
        except FieldNotPresentError:
            return []

    def get_paths_and_objects(
        self,
        doc: dict[str, Any],
        field_path: str | list[Union[str, QueryParser]],
        fallback_variables: dict[str, list[Any]],
        path_prefix: str = "",
    ) -> dict[str, Any]:
        """
        Retrieves a dictionary of absolute paths and values
        on these paths according to

        """
        parsed_path = self.ensure_field_path(field_path)
        resolved_variables = self.__cast_vars_to_sets(
            self.__populate_variables(
                doc,
                fallback_variables,
                parsed_path,
                allow_fail=True,
                prefer_fallback=True,
            )
        )
        ans = {}

        def extend_ans(value: Any, path: str):
            ans[path] = value

        self._run_on_path(
            doc,
            parsed_path,
            resolved_variables,
            path_prefix,
            extend_ans,
            False,
            False,
        )
        return ans

    def get_mutable_parents(
        self,
        doc: dict[str, Any],
        field_path: str | list[Union[str, QueryParser]],
        fallback_variables: dict[str, Any] | None = None,
        create_nonexistent: bool = False,
    ) -> list[Any]:
        """
        Fetches the parent of the last expression. Useful for document mutations.
        Does not resolve variables on its own.
        """
        path = self.ensure_field_path(field_path)
        if create_nonexistent:
            # create parents
            self.get_objects(doc, path, fallback_variables, create_nonexistent)
        # fetch parents
        return self.get_objects(doc, path[:-1], fallback_variables, create_nonexistent)



import sys
from pathlib import Path

from sbomgrader.grade.cookbook_bundles import CookbookBundle
from sbomgrader.grade.cookbooks import Cookbook


def select_cookbook_bundle(cookbooks: list[str]) -> CookbookBundle:
    cookbook_bundles = []
    default_cookbooks = Cookbook.load_all_defaults()
    cookbook_bundle = CookbookBundle([])
    for cookbook in cookbooks:
        cookbook_obj = next(
            filter(lambda x: x.name == cookbook, default_cookbooks), None
        )
        if cookbook_obj:
            # It's a default cookbook name
            cookbook_bundle += cookbook_obj
            continue
        cookbook = Path(cookbook)
        if cookbook.is_dir():
            cookbook_bundle += CookbookBundle.from_directory(cookbook)
            if not cookbook_bundle.cookbooks:
                print(
                    f"Could not find any cookbooks in directory {cookbook.absolute()}",
                    file=sys.stderr,
                )
        elif cookbook.is_file() and (
            cookbook.name.endswith(".yml") or cookbook.name.endswith(".yaml")
        ):
            cookbook_bundles.append(CookbookBundle([Cookbook.from_file(cookbook)]))
        else:
            print(f"Could not find cookbook {cookbook.absolute()}", file=sys.stderr)

    for cb in cookbook_bundles:
        cookbook_bundle += cb
    return cookbook_bundle



from typing import Any
from datetime import datetime

from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.json_like_dict_parser import JsonLikeDictParser
from spdx_tools.spdx.validation.document_validator import validate_full_spdx_document


def validate_schema(doc: dict[str, Any]):
    try:
        JsonLikeDictParser().parse(doc)
    except SPDXParsingError as e:
        raise AssertionError(*e.args)


def full_validation(doc: dict[str, Any]):
    try:
        document = JsonLikeDictParser().parse(doc)
    except SPDXParsingError as e:
        raise AssertionError(*e.args)
    validations = validate_full_spdx_document(document, "SPDX-2.3")
    if validations:
        raise AssertionError(
            *[validation.validation_message for validation in validations]
        )


def check_datetime(doc: str):
    try:
        datetime.fromisoformat(doc)
    except:
        raise AssertionError("Cannot parse ISO datetime.")


def package_relationships(doc: dict[str, Any]):
    package_spdxids = {package["SPDXID"] for package in doc.get("packages", [])}
    for relationship in doc.get("relationships", []):
        references = relationship.get("spdxElementId"), relationship.get(
            "relatedSpdxElement"
        )
        for r in references:
            if r in package_spdxids:
                package_spdxids.remove(r)
        if not package_spdxids:
            break
    assert (
        not package_spdxids
    ), f"Not all packages are referenced in relationships: {package_spdxids}"



from copy import deepcopy
from typing import Union

import pytest

from sbomgrader.core.enums import QueryType
from sbomgrader.core.field_resolve import (
    PathParser,
    QueryParser,
    FieldResolver,
    Query,
    Variable,
)


@pytest.mark.parametrize(
    ["path", "relative_path", "output"],
    [
        ("foo", "", ["foo"]),
        ("foo.bar", "", ["foo", "bar"]),
        ("foo[1]bar", "", ["foo", QueryParser("1"), "bar"]),
        ("@.foo[1]bar", "spam.ham", ["spam", "ham", "foo", QueryParser("1"), "bar"]),
        (
            "foo[@]bar[@]spam",
            "foo[4]bar[2]ham[7]",
            ["foo", QueryParser(4), "bar", QueryParser("2"), "spam"],
        ),
    ],
)
def test_path_parser(path: str, relative_path: str, output: list[Union]):
    assert PathParser(path).parse(relative_path) == output


@pytest.mark.parametrize(
    ["path"],
    [
        ("hello.foo[bar=1,spam!=2]ham",),
        ("hello . foo [bar =1, spam != 2 ] ham",),
        (".hello.foo.[.bar.=1,.spam.!= 2 ].ham",),
        (" . hello . foo . [ . bar . =1, . spam . != 2] . ham",),
        ("..hello..foo..[..bar..=1,..spam..!=2 ]..ham",),
    ],
)
def test_path_parser_unambiguity(path: str):
    assert PathParser(path).parse() == [
        "hello",
        "foo",
        QueryParser("bar=1,spam!=2"),
        "ham",
    ]


@pytest.mark.parametrize(
    ["query_str"],
    [
        ("bar=1,spam!=2",),
        ("bar =1, spam !=2",),
        (".bar.=1,.spam.!=2",),
        (" . bar . =1, . spam . !=2",),
        ("..bar..=1,..spam..!=2",),
    ],
)
def test_query_parser_unambiguity(query_str: str):
    assert QueryParser(query_str).parse() == [
        Query(QueryType.EQ, value="1", field_path=PathParser("bar")),
        Query(QueryType.NEQ, value="2", field_path=PathParser("spam")),
    ]


@pytest.mark.parametrize(
    ["path", "document", "output"],
    [
        (
            "dependencies[provides[&]%=]",
            {
                "dependencies": [
                    {
                        "ref": "foo",
                        "provides": ["bar"],
                        "dependsOn": [],
                    },
                    {
                        "ref": "spam",
                        "provides": [],
                        "dependsOn": ["ham"],
                    },
                ]
            },
            [  # Output is always a list
                {
                    "ref": "foo",
                    "provides": ["bar"],
                    "dependsOn": [],
                }
            ],
        ),
        (
            "foo[bar=FIELD_NOT_PRESENT]",
            {"foo": [{"bar": 1}, {"spam": 2}]},
            [{"spam": 2}],
        ),
    ],
)
def test_fetch_fields(path, document, output):
    resolver = FieldResolver({})
    assert output == resolver.get_objects(document, path)


@pytest.mark.parametrize(
    ["path", "expected"],
    [
        (".=.", [Query(QueryType.EQ, ".", PathParser("."))]),
        (
            "?.purl%=pkg:rpm/,?.purl%arch=src",
            [
                Query(QueryType.STARTSWITH, "pkg:rpm/", PathParser("?.purl")),
                Query(QueryType.CONTAINS, "arch=src", PathParser("?.purl")),
            ],
        ),
        (
            "externalRefs[referenceType=purl]referenceLocator%=pkg:generic/",
            [
                Query(
                    QueryType.STARTSWITH,
                    "pkg:generic/",
                    PathParser("externalRefs[referenceType=purl]referenceLocator"),
                )
            ],
        ),
    ],
)
def test_query_parser(path: str, expected: list[Query]):
    assert QueryParser(path).parse() == expected


@pytest.mark.parametrize(
    ["path", "variables", "original_doc", "expected_doc"],
    [
        (".foo.bar", {}, {}, {"foo": {}}),
        ("foo[|]", {}, {}, {"foo": []}),
        ("?.foo.?.bar", {}, {}, {"foo": {}}),
        (
            "foo[|]?.bar.spam",
            {},
            {"foo": [{"ham": 1}]},
            {"foo": [{"ham": 1, "bar": {}}]},
        ),
        (
            "foo[|]?.bar[|].?.spam.?.ham",
            {},
            {"foo": [{"bar": [{"hi": 1}]}]},
            {"foo": [{"bar": [{"hi": 1, "spam": {}}]}]},
        ),
    ],
)
def test_create_nonexistent(
    path: str, variables: dict[str, Variable], original_doc, expected_doc
):
    """Tests if non-present fields are correctly created"""
    resolver = FieldResolver(variables)
    testing_doc = deepcopy(original_doc)
    resolver.get_objects(testing_doc, path, {}, True)
    assert expected_doc == testing_doc



import datetime
import json
from json import JSONDecodeError
from pathlib import Path
from typing import Any, Literal

import jinja2
import yaml
from jsonschema import validate
from yaml import YAMLError

from sbomgrader.core.cached_python_loader import PythonLoader
from sbomgrader.core.definitions import FIELD_NOT_PRESENT, TIME_ISO_FORMAT_STRING
from sbomgrader.core.enums import Grade
from sbomgrader import __version__ as version


def is_mapping(file: str | Path) -> bool:
    name = file if isinstance(file, str) else file.name
    return name.endswith(".json") or name.endswith(".yml") or name.endswith(".yaml")


def get_mapping(
    schema: str | Path, validation_schema: str | Path | None = None
) -> dict | None:
    """
    Load a mapping from a JSON/YAML file, optionally validate it with a JSONSchema.
    This function can also load the mapping from a string.

    :argument schema: The mapping to load. Can hold the string-serialized mapping
    or a path to a JSON or YAML file.
    :argument validation_schema: The JSONSchema used for validation of the first mapping.
    """
    doc = {}
    try:
        if isinstance(schema, str):
            if schema.startswith("---"):
                doc = yaml.safe_load(schema)
            elif schema.startswith("{"):
                doc = json.loads(schema)
            else:
                schema = Path(schema)
    except (JSONDecodeError, YAMLError):
        # Some people maybe really do name
        # their files with names starting with '{'
        # or '---'
        schema = Path(schema)

    if isinstance(schema, Path):
        if not schema.exists() or not is_mapping(schema):
            return None
        with open(schema) as stream:
            if schema.name.endswith(".json"):
                doc = json.load(stream)
            elif schema.name.endswith(".yml") or schema.name.endswith(".yaml"):
                doc = yaml.safe_load(stream)

    if not doc:
        raise ValueError(f"Invalid mapping: '{schema}'.")
    if validation_schema:
        validate(doc, get_mapping(validation_schema))
    return doc


def get_path_to_implementations(schema_path: str | Path) -> Path:
    """Get a relative path to the module containing test implementation functions of this Rule Set."""
    if isinstance(schema_path, str):
        schema_path = Path(schema_path)
    return schema_path.parent / "implementations" / schema_path.name.rsplit(".", 1)[0]


def get_path_to_var_transformers(schema_path: str | Path) -> Path:
    """Get a relative path to the module containing transformer functions of this Translation Map."""
    if isinstance(schema_path, str):
        schema_path = Path(schema_path)
    return schema_path.parent / "transformers" / schema_path.name.split(".", 1)[0]


def get_path_to_preprocessing(schema_path: str | Path) -> Path:
    """Get a relative path to the module containing preprocessing functions of this Translation Map."""
    if isinstance(schema_path, str):
        schema_path = Path(schema_path)
    return schema_path.parent / "preprocessing" / schema_path.name.split(".", 1)[0]


def get_path_to_postprocessing(schema_path: str | Path) -> Path:
    """Get a relative path to the module containing postprocessing functions of this Translation Map."""
    if isinstance(schema_path, str):
        schema_path = Path(schema_path)
    return schema_path.parent / "postprocessing" / schema_path.name.split(".", 1)[0]


def get_path_to_module(
    schema_path: str | Path,
    kind: Literal["Transformer", "Preprocessing", "Postprocessing"],
    first_or_second: Literal["first", "second"],
    sbom_format: "sbomgrader.core.formats.SBOMFormat,",
):
    """
    Get a path to a module related to the translation map located
    in the file at `schema_path`.
    This loads either transformers, preprocessors or postprocessors.
    :argument schema_path: Path to Translation Map
    :argument kind: Type of module file (what type of function is searched for)
    :argument first_or_second: Do we search for the file related to `first`
    or `second` SBOM Format?
    :argument sbom_format: What is the SBOM Format this module relates to?
    """
    map_kind_to_module_function = {
        "Transformer": get_path_to_var_transformers,
        "Preprocessing": get_path_to_preprocessing,
        "Postprocessing": get_path_to_postprocessing,
    }
    mod_func = map_kind_to_module_function.get(kind)
    if not mod_func:
        raise ValueError(f"Wrong kind value: {kind}")
    mod_dir = mod_func(schema_path)
    file = None
    for filename in (
        f"{first_or_second}.py",
        f"{sbom_format.value}.py",
    ):
        file_path = mod_dir / filename
        if file_path.exists():
            file = file_path
            break
    return file


def validation_passed(validation_grade: Grade, minimal_grade: Grade) -> bool:
    """Did the SBOM get a good enough grade?"""
    # minimal is less than or equal to validation
    return Grade.compare(validation_grade, minimal_grade) < 1


def create_jinja_env(transformer_file: Path | None = None) -> jinja2.Environment:
    """Creates a Jinja2 environment with additional filters. Used in Translation Maps."""
    env = jinja2.Environment()
    env.globals["DATETIME_NOW"] = datetime.datetime.now(datetime.UTC).strftime(
        TIME_ISO_FORMAT_STRING
    )
    env.globals["SBOMGRADER_SIGNATURE"] = f"SBOMGrader {version}"

    def unwrap(input_list: list[Any]) -> Any:
        """Return the first element of a list."""
        try:
            return next(iter(input_list), "")
        except TypeError:
            return ""

    def sliced(
        input_list: list[Any] | str, start: int = 0, end: int = None
    ) -> list[Any] | str:
        """Return a slice of a list or a string."""
        if not isinstance(input_list, list) and not isinstance(input_list, str):
            return []
        return input_list[start:end]

    def unify(first: list[Any], *other: list[Any]) -> list[Any]:
        """Return union"""
        ans = []
        if isinstance(first, list):
            ans.extend(first)
        for o in other:
            if isinstance(o, list):
                ans.extend(o)

        return ans

    def fallback(first: list[Any], *other: list[Any]) -> list[Any]:
        """Return the first non-empty variable value (non-empty list)."""
        if first and first is not FIELD_NOT_PRESENT:
            return first
        for o in other:
            if o and o is not FIELD_NOT_PRESENT:
                return o
        return []

    env.filters["unwrap"] = unwrap
    env.filters["slice"] = sliced
    env.filters["fallback"] = fallback
    env.filters["unify"] = unify
    if transformer_file and transformer_file.exists():

        def func(item: Any, name: str, **kwargs) -> Any:
            python_loader = PythonLoader(transformer_file)
            return python_loader.load_func(name)(item, **kwargs)

        env.filters["func"] = func
    return env



import re
from pathlib import Path
from typing import Sized


ROOT_DIR: Path = Path(__file__).parent.parent
RULESET_DIR = ROOT_DIR / "rulesets"
COOKBOOKS_DIR = ROOT_DIR / "cookbooks"
IMPLEMENTATION_DIR_NAME = "specification_rules"
RULESET_VALIDATION_SCHEMA_PATH = ROOT_DIR / "rulesets" / "schema" / "rule_schema.yml"
COOKBOOK_VALIDATION_SCHEMA_PATH = (
    ROOT_DIR / "cookbooks" / "schema" / "cookbook_schema.yml"
)
TRANSLATION_MAP_DIR = ROOT_DIR / "translation_maps"
TRANSLATION_MAP_VALIDATION_SCHEMA_PATH = (
    TRANSLATION_MAP_DIR / "schema" / "translation_map_schema.yml"
)
FORMAT_FILE_PATH = ROOT_DIR / "formats" / "formats.yml"
FORMAT_VALIDATION_SCHEMA_PATH = ROOT_DIR / "formats" / "schema" / "formats_schema.yml"
COOKBOOK_EXTENSIONS = {".yml", ".yaml"}

TIME_ISO_FORMAT_STRING = "%Y-%m-%dT%H:%M:%SZ"

MAX_ITEM_PREVIEW_LENGTH = 50
START_PREVIEW_CHARS = 25
END_PREVIEW_CHARS = 20
VAR_REF_REGEX = r"\${(?P<var_id>[^}]+)}"


class __FieldNotPresent:
    string_repr = "FIELD_NOT_PRESENT"

    def __repr__(self):
        return "Field not present."

    def get(self, *_):
        return self


FIELD_NOT_PRESENT = __FieldNotPresent()


class FieldNotPresentError(ValueError):
    pass


operation_map = {
    "eq": lambda expected, actual: expected == actual,
    "neq": lambda expected, actual: expected != actual,
    "in": lambda expected, actual: actual in expected,
    "not_in": lambda expected, actual: actual not in expected,
    "str_startswith": lambda expected, actual: isinstance(actual, str)
    and actual.startswith(expected),
    "str_endswith": lambda expected, actual: isinstance(actual, str)
    and actual.endswith(expected),
    "str_contains": lambda expected, actual: isinstance(actual, str)
    and expected in actual,
    "str_matches_regex": lambda expected, actual: isinstance(actual, str)
    and bool(re.match(expected, actual)),
    "length_eq": lambda expected, actual: isinstance(actual, Sized)
    and len(actual) == expected,
    "length_gt": lambda expected, actual: isinstance(actual, Sized)
    and len(actual) > expected,
    "length_lt": lambda expected, actual: isinstance(actual, Sized)
    and len(actual) < expected,
    "func_name": None,
}



from typing import Any

from sbomgrader.core.field_resolve import FieldResolver


def clone_main_element(doc: dict[str, Any]):
    main_component = doc.get("metadata", {}).get("component", {})
    if not main_component:
        raise ValueError(
            "Please provide an SBOM with a valid metadata.component field."
        )
    if main_component.get("bom-ref"):
        return
    applicable_identifiers = [
        x for x in ("purl", "cpe", "identity", "name") if x in main_component
    ]
    assert (
        applicable_identifiers
    ), "Cannot convert SBOM that does not have any identifiers of the metadata.component field."
    for identifier in applicable_identifiers:
        id_value = main_component.get(identifier)
        for comp in doc.get("components", []):
            if comp.get(identifier) == id_value:
                # This is the main component
                main_component.update(comp)


def ensure_tools_format(doc: dict[str, Any]):
    tools = doc.get("metadata", {}).get("tools")
    if isinstance(tools, list):
        doc["metadata"]["tools"] = {
            "components": [{"name": t, "type": "application"} for t in tools]
        }


def __insert(doc: dict[str, Any], value: Any, path: list[str]) -> None:
    for step in path[:-1]:
        if step not in doc:
            doc[step] = {}
        doc = doc[step]
    doc[path[-1]] = value


def ensure_supplier_format(doc: dict[str, Any]):
    """Makes sure all suppliers are in expected format."""
    if (
        supplier := doc.get("metadata", {}).get("component", {}).get("supplier")
    ) and not doc.get("metadata", {}).get("supplier", {}).get("name"):
        __insert(doc, supplier, ["metadata", "supplier"])
    elif supplier := doc.get("metadata", {}).get("supplier") and not doc.get(
        "metadata", {}
    ).get("component", {}).get("supplier", {}).get("name"):
        __insert(doc, supplier, ["metadata", "component", "supplier"])

    paths = ["metadata", "components[|]"]
    resolver = FieldResolver({})
    for path in paths:
        for element in resolver.get_objects(doc, path):
            if "supplier" not in element:
                continue
            supplier = element["supplier"]
            if isinstance(supplier, str):
                element["supplier"] = {"name": supplier}



from enum import Enum


class ResultType(Enum):
    SUCCESS = "success"
    NOT_PRESENT = "not present"
    FAILED = "failed"
    ERROR = "error"
    SKIPPED = "skipped"
    NOT_IMPLEMENTED = "not implemented"

    @staticmethod
    def get_visual(res_type: "ResultType") -> str:
        mapping = {
            ResultType.SUCCESS: "\N{CHECK MARK}",
            ResultType.FAILED: "\N{CROSS MARK}",
            ResultType.ERROR: "!",
            ResultType.SKIPPED: "\N{RIGHT-SIDE ARC CLOCKWISE ARROW}",
            ResultType.NOT_PRESENT: "??",
            ResultType.NOT_IMPLEMENTED: "?",
        }
        return mapping[res_type]


class RuleForce(Enum):
    MUST = "MUST"
    SHOULD = "SHOULD"
    MAY = "MAY"


class SBOMTime(Enum):
    RELEASE = "release"
    BUILD = "build"
    UNSPECIFIED = None


class Grade(Enum):
    A = "A"
    B = "B"
    C = "C"
    D = "D"
    E = "E"
    F = "F"

    @staticmethod
    def lower(grade: "Grade"):
        if grade is Grade.F:
            return grade.F
        return Grade(chr(ord(grade.value) + 1))

    @staticmethod
    def compare(left: "Grade", right: "Grade") -> int:
        diff = ord(left.value) - ord(right.value)
        if not diff:
            return diff
        return diff // abs(diff)


class SBOMType(Enum):
    IMAGE = "image"
    IMAGE_INDEX = "image_index"
    RPM = "rpm"
    PRODUCT = "product"
    UNKNOWN = "generic"
    UNSPECIFIED = None


class OutputType(Enum):
    YAML = "yaml"
    JSON = "json"
    MARKDOWN = "markdown"
    VISUAL = "visual"


class QueryType(Enum):
    EACH = "&"
    ANY = "|"
    EQ = "="
    NEQ = "!="
    CONTAINS = "%"
    NOT_CONTAINS = "!%"
    STARTSWITH = "%="
    ENDSWITH = "=%"
    RELATIVE = "@"
    INDEX = ""



from pathlib import Path
from typing import Any, Callable

import yaml
from jinja2 import meta

from sbomgrader.core.cached_python_loader import PythonLoader
from sbomgrader.core.definitions import (
    TRANSLATION_MAP_VALIDATION_SCHEMA_PATH,
    FIELD_NOT_PRESENT,
)
from sbomgrader.core.documents import Document
from sbomgrader.core.field_resolve import (
    Variable,
    FieldResolver,
    QueryParser,
)
from sbomgrader.core.formats import (
    SBOMFormat,
    get_fallbacks,
    SBOM_FORMAT_DEFINITION_MAPPING,
)
from sbomgrader.core.utils import (
    get_mapping,
    create_jinja_env,
    get_path_to_module,
)
from sbomgrader.translate.prune import prune, should_remove


class Data:
    """The data to render in the new document."""

    def __init__(
        self,
        template: str,
        variables: dict[str, Variable],
        transformer_path: Path | None = None,
    ):
        self.variables = variables
        self.template = template
        self.field_resolver = FieldResolver(variables)
        self.transformer_path = transformer_path
        self.jinja_env = create_jinja_env(self.transformer_path)
        self._variables_needed_in_template = self._get_vars_for_template()

    def render(
        self,
        whole_doc: Document,
        path_to_instance: str | None,
        instance_value: Any,
        prune_empty: bool = True,
        globally_resolved_variables: dict[str, list[Any]] = None,
    ) -> Any:
        """
        Renders a Jinja2 expression according to variables
        populated from the document.
        """
        globally_resolved_variables = globally_resolved_variables or {}
        already_resolved_vars = {**globally_resolved_variables}
        relative_resolver = FieldResolver(
            {
                var_name: var.without_relative_start
                for var_name, var in self.field_resolver.fully_relative_variables.items()
            }
        )
        already_resolved_vars.update(
            relative_resolver.resolve_variables(
                instance_value,
                already_resolved_variables=already_resolved_vars,
                path_prefix=path_to_instance,
            )
        )
        resolved_variables = self.field_resolver.resolve_variables(
            whole_doc.doc,
            path_to_instance,
            already_resolved_variables=already_resolved_vars,
            variables_needed=self._variables_needed_in_template,
        )
        # Remove invalid values
        for var_name, var_val in resolved_variables.items():
            resolved_variables[var_name] = [
                val for val in var_val if val is not FIELD_NOT_PRESENT
            ]
        data_value = yaml.safe_load(
            self.jinja_env.from_string(self.template).render(**resolved_variables)
        )
        if prune_empty:
            data_value = prune(data_value)
        return data_value

    def _get_vars_for_template(self) -> set[str]:
        parsed_content = self.jinja_env.parse(self.template)
        jinja_vars = meta.find_undeclared_variables(parsed_content)
        return {var for var in jinja_vars if var not in self.jinja_env.globals}


class Chunk:
    """A piece of information represented in 2 SBOM formats."""

    def __init__(
        self,
        name: str,
        first_format: SBOMFormat,
        second_format: SBOMFormat,
        first_data: Data | None,
        second_data: Data | None,
        first_field_path: str,
        second_field_path: str,
        first_variables: dict[str, Variable] = None,
        second_variables: dict[str, Variable] = None,
    ):
        self.name = name
        self.first_format = first_format
        self.second_format = second_format
        self.first_data = first_data
        self.second_data = second_data
        self.first_field_path = first_field_path
        self.second_field_path = second_field_path
        self.first_variables = first_variables or {}
        self.second_variables = second_variables or {}
        self.first_resolver = FieldResolver(self.first_variables)
        self.second_resolver = FieldResolver(self.second_variables)

    def _first_or_second(self, sbom_format: SBOMFormat) -> str:
        if sbom_format == self.first_format or self.first_format in get_fallbacks(
            sbom_format
        ):
            return "first_"
        if sbom_format == self.second_format or self.second_format in get_fallbacks(
            sbom_format
        ):
            return "second_"
        raise ValueError(f"This map does not support format {sbom_format}!")

    def _other(self, sbom_format: SBOMFormat) -> SBOMFormat:
        if sbom_format == self.first_format:
            return self.second_format
        if sbom_format == self.second_format:
            return self.first_format
        raise ValueError(f"This map does not support format {sbom_format}!")

    def data_for(self, sbom_format: SBOMFormat) -> Data:
        return getattr(self, f"{self._first_or_second(sbom_format)}data")

    def field_path_for(self, sbom_format: SBOMFormat) -> str | None:
        return getattr(self, f"{self._first_or_second(sbom_format)}field_path")

    def resolver_for(self, sbom_format: SBOMFormat) -> FieldResolver:
        return getattr(self, f"{self._first_or_second(sbom_format)}resolver")

    def occurrences(
        self, doc: Document, fallback_variables: dict[str, Any] = None
    ) -> dict[str, Any]:
        """Returns a list of objects and string fieldPaths where the element occurs."""
        fallback_variables = fallback_variables or {}
        resolver = self.resolver_for(doc.sbom_format)
        return resolver.get_paths_and_objects(
            doc.doc, self.field_path_for(doc.sbom_format), fallback_variables
        )

    @staticmethod
    def __mutate_obj_by_inserting(
        parent_obj: Any,
        value_to_insert: Any,
        last_step_field_path: str | QueryParser | None,
    ) -> None:
        if last_step_field_path is None:
            # This is at the root level
            parent_obj.update(value_to_insert)
        if isinstance(last_step_field_path, str):
            # We want to add to a dict
            parent_obj[last_step_field_path] = value_to_insert
        elif isinstance(last_step_field_path, QueryParser):
            # We want to add to a list
            if isinstance(value_to_insert, list):
                parent_obj.extend(value_to_insert)
            else:
                parent_obj.append(value_to_insert)

    def convert_and_add(
        self,
        orig_doc: Document,
        new_doc: dict[str, Any],
        globally_resolved_variables: dict[str, list[Any]] = None,
    ) -> None:
        """Mutates the new_doc with the occurrences of this chunk."""
        convert_from = orig_doc.sbom_format
        if convert_from not in {self.first_format, self.second_format}:
            fallbacks = get_fallbacks(orig_doc.sbom_format)
            if self.first_format in fallbacks:
                convert_from = self.first_format
            elif self.second_format in fallbacks:
                convert_from = self.second_format
        convert_to = (
            self.first_format
            if self.first_format != convert_from
            else self.second_format
        )
        relevant_data = self.data_for(convert_to)
        if not relevant_data:
            # This chunk does not specify anything for this direction
            return
        globally_resolved_variables = globally_resolved_variables or {}

        source_resolver = self.resolver_for(convert_from)
        chunk_based_absolute_vars = source_resolver.resolve_variables(orig_doc.doc)
        global_vars = {**globally_resolved_variables, **chunk_based_absolute_vars}

        # Resolve all info about the point where to insert data -- once
        appender_resolver = self.resolver_for(convert_to)
        append_path = appender_resolver.ensure_field_path(
            self.field_path_for(convert_to)
        )
        last_insert_step = next(iter(append_path[-1:]), None)
        mutable_parents = appender_resolver.get_mutable_parents(
            new_doc, append_path, create_nonexistent=True
        )

        for occurrence_path, occurrence_value in self.occurrences(
            orig_doc, globally_resolved_variables
        ).items():
            rendered_data = relevant_data.render(
                orig_doc,
                occurrence_path,
                occurrence_value,
                globally_resolved_variables=global_vars,
            )
            if not should_remove(rendered_data):
                for mutable_parent in mutable_parents:
                    self.__mutate_obj_by_inserting(
                        mutable_parent, rendered_data, last_insert_step
                    )


class TranslationMap:
    """This objects transforms SBOMs between formats."""

    def __init__(
        self,
        first: SBOMFormat,
        second: SBOMFormat,
        chunks: list[Chunk],
        first_variables: dict[str, Variable],
        second_variables: dict[str, Variable],
        preprocessing_funcs: dict[SBOMFormat, list[Callable]] = None,
        postprocessing_funcs: dict[SBOMFormat, list[Callable]] = None,
    ):
        self.first = first
        self.second = second
        self.chunks = chunks
        self.first_variables = first_variables or {}
        self.second_variables = second_variables or {}
        self.preprocessing_funcs: dict[SBOMFormat, list[Callable[[dict], Any]]] = (
            preprocessing_funcs or {}
        )
        self.postprocessing_funcs: dict[
            SBOMFormat, list[Callable[[dict, dict], Any]]
        ] = (postprocessing_funcs or {})

    @staticmethod
    def from_file(file: str | Path) -> "TranslationMap":
        """Load the Translation Map from a file."""
        schema_dict = get_mapping(file, TRANSLATION_MAP_VALIDATION_SCHEMA_PATH)

        first = SBOMFormat(schema_dict["first"])
        second = SBOMFormat(schema_dict["second"])

        first_glob_var = schema_dict.get("firstVariables")
        second_glob_var = schema_dict.get("secondVariables")

        first_transformer_file = get_path_to_module(file, "Transformer", "first", first)
        first_glob_var_initialized = Variable.from_schema(first_glob_var)
        second_transformer_file = get_path_to_module(
            file, "Transformer", "second", second
        )
        second_glob_var_initialized = Variable.from_schema(second_glob_var)

        chunks = []
        for chunk_dict in schema_dict["chunks"]:
            name = chunk_dict["name"]

            first_field_path = chunk_dict.get("firstFieldPath")
            second_field_path = chunk_dict.get("secondFieldPath")
            first_variables = Variable.from_schema(chunk_dict.get("firstVariables"))
            second_variables = Variable.from_schema(chunk_dict.get("secondVariables"))

            first_vars = {**first_glob_var_initialized}
            first_vars.update(first_variables)

            second_vars = {**second_glob_var_initialized}
            second_vars.update(second_variables)

            if first_data_dict := chunk_dict.get("firstData"):
                first_data = Data(first_data_dict, second_vars, second_transformer_file)
            else:
                first_data = None
            if second_data_dict := chunk_dict.get("secondData"):
                second_data = Data(second_data_dict, first_vars, first_transformer_file)
            else:
                second_data = None
            chunk = Chunk(
                name,
                first,
                second,
                first_data,
                second_data,
                first_field_path,
                second_field_path,
                first_vars,
                second_vars,
            )
            chunks.append(chunk)

        preprocessing_dict = {}
        postprocessing_dict = {}
        for dict_of_functions, kind in (
            (preprocessing_dict, "Preprocessing"),
            (postprocessing_dict, "Postprocessing"),
        ):
            # Load both preprocessing and postprocessing functions
            for first_or_second, sbom_format in ("first", first), ("second", second):
                # Load both the functions for the first and the second format
                required_funcs = schema_dict.get(f"{first_or_second}{kind}", [])
                if not required_funcs:
                    # There are no functions required by the TranslationMap
                    continue
                dict_of_functions[sbom_format] = []
                py_file = get_path_to_module(file, kind, first_or_second, sbom_format)
                python_loader = PythonLoader(py_file)
                for func_name in required_funcs:
                    dict_of_functions[sbom_format].append(
                        python_loader.load_func(func_name)
                    )

        return TranslationMap(
            first,
            second,
            chunks,
            first_glob_var_initialized,
            second_glob_var_initialized,
            preprocessing_dict,
            postprocessing_dict,
        )

    def _output_format(self, doc: Document) -> SBOMFormat:
        for form in self.first, self.second:
            if doc.sbom_format is not form and not any(
                doc.sbom_format == fallback for fallback in get_fallbacks(form)
            ):
                return form

    def _input_format(self, doc: Document) -> SBOMFormat:
        for form in self.first, self.second:
            if doc.sbom_format is form:
                return form
        for form in self.first, self.second:
            if doc.sbom_format in get_fallbacks(form):
                return form
        raise ValueError(f"Cannot do anything with this format: {doc.sbom_format}")

    def convert(
        self, sbom: Document, override_format: SBOMFormat | None = None
    ) -> Document:
        """
        Converts document to the specified format.
        :argument sbom: Sbom document to convert.
        :argument override_format: Specify which is the output format.
        If omitted, the format is chosen from values self.first or
        self.second. The value not associated with input document will be used.
        """
        new_data = {}
        assert sbom.sbom_format in (
            self.first,
            self.second,
        ) or any(
            fallback
            in (
                self.first,
                self.second,
            )
            for fallback in sbom.sbom_format_fallback
        ), f"This map cannot convert from {sbom.sbom_format}."
        # Preprocess
        sbom_dict = sbom.doc
        for preprocessing_func in self.preprocessing_funcs.get(
            self._input_format(sbom), []
        ):
            res = preprocessing_func(sbom_dict)
            if res:
                sbom_dict = res
        # Finish preprocessing
        sbom = Document(sbom_dict)

        # Load global vars
        variable_definitions = (
            self.first_variables
            if self._input_format(sbom) == self.first
            else self.second_variables
        )
        globally_loaded_variables = FieldResolver(
            variable_definitions
        ).resolve_variables(sbom.doc)

        # Conversion
        for chunk in self.chunks:
            chunk.convert_and_add(sbom, new_data, globally_loaded_variables)
        # Postprocess
        for postprocessing_func in self.postprocessing_funcs.get(
            self._output_format(sbom), []
        ):
            res = postprocessing_func(sbom.doc, new_data)
            if res:
                # If the function returns anything, make it the new data output.
                # Assume mutations were performed in-place otherwise.
                new_data = res
        if override_format is not None:
            new_data.update(SBOM_FORMAT_DEFINITION_MAPPING[override_format])
        return Document(new_data)

    def is_exact_map(self, from_: SBOMFormat, to: SBOMFormat) -> bool:
        """Determine if this map converts between these two formats."""
        return ((from_ is self.first) and (to is self.second)) or (
            (from_ is self.second) and (to is self.first)
        )

    def is_suitable_map(self, from_: SBOMFormat, to: SBOMFormat) -> bool:
        """Determine if the map is able to convert between formats including fallbacks."""
        if self.is_exact_map(from_, to):
            return True
        from_fallbacks = get_fallbacks(from_)
        from_fallbacks.add(from_)
        to_fallbacks = get_fallbacks(to)
        to_fallbacks.add(to)
        return (self.first in from_fallbacks and self.second in to_fallbacks) or (
            self.first in to_fallbacks and self.second in from_fallbacks
        )



import json
from functools import cached_property
from pathlib import Path
from typing import Any

from sbomgrader.core.enums import SBOMType
from sbomgrader.core.formats import (
    SBOM_FORMAT_DEFINITION_MAPPING,
    SBOMFormat,
    get_fallbacks,
)
from sbomgrader.core.utils import get_mapping


class Document:
    def __init__(self, document_dict: dict[str, Any]):
        self._doc = document_dict

    @cached_property
    def sbom_format(self) -> SBOMFormat:
        for item in SBOMFormat:
            field_to_check = SBOM_FORMAT_DEFINITION_MAPPING[item]

            if all(
                self._doc.get(key) == value for key, value in field_to_check.items()
            ):
                return item
        raise NotImplementedError("Document is in an unsupported standard.")

    @property
    def sbom_format_fallback(self) -> set[SBOMFormat]:
        return get_fallbacks(self.sbom_format)

    @property
    def sbom_type(self) -> "SBOMType":
        if self.sbom_format is SBOMFormat.SPDX23 or self.sbom_format in get_fallbacks(
            SBOMFormat.SPDX23
        ):
            relationships = self._doc.get("relationships", [])
            main_relationships = [
                relationship
                for relationship in relationships
                if relationship["spdxElementId"] == "SPDXRef-DOCUMENT"
                and relationship["relationshipType"] == "DESCRIBES"
            ]
            if len(main_relationships) > 1:
                raise ValueError(
                    "Cannot determine single SBOMType from multi-sbom. Try separating docs first."
                )
            main_relationship = main_relationships[0]
            main_spdxid = main_relationship["relatedSpdxElement"]
            first_degree_relationships = [
                relationship
                for relationship in relationships
                if (
                    relationship["spdxElementId"] == main_spdxid
                    or relationship["relatedSpdxElement"] == main_spdxid
                )
                and relationship != main_relationship
            ]
            if all(
                relationship["relationshipType"] == "VARIANT_OF"
                for relationship in first_degree_relationships
            ):
                return SBOMType.IMAGE_INDEX
            if all(
                relationship["relationshipType"]
                in {"DESCENDANT_OF", "CONTAINS", "BUILD_TOOL_OF"}
                for relationship in first_degree_relationships
            ):
                return SBOMType.IMAGE
            if all(
                relationship["relationshipType"] in {"GENERATED_FROM", "CONTAINS"}
                for relationship in first_degree_relationships
            ):
                return SBOMType.RPM

            def sort_relationship_key(relationship: dict):
                return "".join(sorted(relationship.values()))

            if sorted(
                first_degree_relationships + main_relationships,
                key=sort_relationship_key,
            ) == sorted(relationships, key=sort_relationship_key):
                return SBOMType.PRODUCT
            return SBOMType.UNKNOWN
        elif (
            self.sbom_format is SBOMFormat.CYCLONEDX16
            or self.sbom_format in get_fallbacks(SBOMFormat.CYCLONEDX16)
        ):
            if self._doc.get("metadata", {}).get("component", {}).get("type") in {
                "operating-system"
            }:
                return SBOMType.PRODUCT
            return SBOMType.UNKNOWN
        else:
            raise NotImplementedError()

    @property
    def doc(self):
        return self._doc

    @property
    def json_dump(self) -> str:
        return json.dumps(self._doc, indent=4)

    @staticmethod
    def from_file(path_to_file: str | Path) -> "Document":
        path_to_file = Path(path_to_file)
        mapping = get_mapping(path_to_file)
        if not mapping:
            raise ValueError(
                f"It seems that file {path_to_file.absolute()} does not contain a valid mapping."
                f"Please make sure a valid json or yaml file is provided."
            )
        return Document(get_mapping(path_to_file))



import sys
from argparse import ArgumentParser, Namespace
from dataclasses import dataclass
from pathlib import Path

import click
from rich.console import Console
from rich.markdown import Markdown

from sbomgrader.core.formats import SBOMFormat
from sbomgrader.grade.choose_cookbooks import select_cookbook_bundle
from sbomgrader.grade.cookbook_bundles import CookbookBundle
from sbomgrader.grade.cookbooks import Cookbook
from sbomgrader.core.documents import Document
from sbomgrader.core.enums import Grade, SBOMTime, OutputType, SBOMType
from sbomgrader.core.utils import get_mapping, validation_passed
from sbomgrader.translate.choose_map import choose_map, get_all_map_list_markdown
from sbomgrader.translate.translation_map import TranslationMap


@dataclass
class GradeConfig:
    input_file: Path | str
    cookbook_references: list[str]
    content_type: SBOMType
    sbom_type: SBOMTime
    passing_grade: Grade
    output_type: OutputType

    @staticmethod
    def from_args(args: Namespace) -> "GradeConfig":
        return GradeConfig(
            input_file=args.input,
            cookbook_references=args.cookbook or [],
            content_type=args.content_type,
            sbom_type=args.sbom_type,
            passing_grade=args.passing_grade,
            output_type=args.output,
        )


def create_grade_parser(parser: ArgumentParser):
    parser.add_argument(
        "input",
        type=str,
        help="SBOM File to grade. Currently supports JSON.",
    )
    parser.add_argument(
        "--cookbook",
        "-c",
        action="append",
        type=str,
        help="Cookbooks to use for validation. "
        "Might reference default cookbook, directories or files. "
        "Only files with '.yml' or '.yaml' extensions are taken into account if files or directories are specified.",
    )
    parser.add_argument(
        "--content-type",
        "-ct",
        choices=[v.value for v in SBOMType if v is not SBOMType.UNSPECIFIED],
        default=SBOMType.UNSPECIFIED.value,
        help="Specify SBOM content type. Ignored if cookbooks argument is specified.",
    )
    parser.add_argument(
        "--sbom-type",
        "-st",
        choices=[v.value for v in SBOMTime if v is not SBOMTime.UNSPECIFIED],
        default=None,
        help="If using the standard validation, specify which SBOM type (by time) is being validated. "
        "Ignored if cookbooks argument is specified.",
    )
    parser.add_argument(
        "--passing-grade",
        "-g",
        choices=[v.value for v in Grade],
        default=Grade.B.value,
        help="Minimal passing grade. Default is B.",
    )
    parser.add_argument(
        "--output",
        "-o",
        choices=[v.value for v in OutputType],
        default=OutputType.VISUAL.value,
        help="Specify the output format.",
    )


def grade(config: GradeConfig) -> None:
    console = Console()

    doc = Document(get_mapping(config.input_file))

    if config.cookbook_references:
        cookbook_bundle = select_cookbook_bundle(config.cookbook_references)
        if not cookbook_bundle.cookbooks:
            print("No cookbook(s) could be found.", file=sys.stderr)
            exit(1)
    else:
        # Cookbooks weren't specified, using defaults
        type_ = SBOMType(config.content_type)
        if type_ is SBOMType.UNSPECIFIED:
            type_ = doc.sbom_type
        cookbook_bundle = CookbookBundle.for_document_type(
            type_, SBOMTime(config.sbom_type)
        )

    result = cookbook_bundle(doc)

    output_type = OutputType(config.output_type)
    if output_type is OutputType.VISUAL:
        markdown = Markdown(result.output(output_type))
        console.print(markdown)
    else:
        console.print(result.output(output_type))
    if validation_passed(result.grade, Grade(config.passing_grade)):
        exit(0)
    exit(1)


@dataclass
class ConvertConfig:
    input_file: Path
    output_format: SBOMFormat
    custom_maps: list[Path]

    @staticmethod
    def from_args(args: Namespace) -> "ConvertConfig":
        return ConvertConfig(
            input_file=args.input,
            output_format=args.output_format,
            custom_maps=args.custom_map or [],
        )


def create_convert_parser(parser: ArgumentParser):
    parser.add_argument(
        "input",
        type=str,
        help="SBOM File to convert. Currently supports JSON.",
    )
    parser.add_argument(
        "--output-format",
        "-f",
        choices=[v.value for v in SBOMFormat],
        required=True,
        help="SBOM format to create.",
    )
    parser.add_argument(
        "--custom-map",
        "-m",
        type=str,
        help="Custom translation map file.",
        action="append",
    )


def convert(config: ConvertConfig) -> None:

    custom_maps = [TranslationMap.from_file(f) for f in config.custom_maps]

    inp_file = config.input_file
    doc = Document(get_mapping(inp_file))

    target_format = SBOMFormat(config.output_format)

    t_map = choose_map(doc, target_format, *custom_maps)
    print(t_map.convert(doc, target_format).json_dump)
    exit(0)


@dataclass
class ListConfig:
    maps: bool
    cookbooks: bool

    @staticmethod
    def from_args(args: Namespace) -> "ListConfig":
        return ListConfig(maps=args.maps, cookbooks=args.cookbooks)


def create_list_parser(parser: ArgumentParser):
    mutually_exclusive = parser.add_mutually_exclusive_group(required=True)
    mutually_exclusive.add_argument(
        "--maps",
        "-m",
        action="store_true",
        default=False,
        help="List available default translation maps and exit.",
    )
    mutually_exclusive.add_argument(
        "--cookbooks",
        "-c",
        action="store_true",
        default=False,
        help="List available default cookbooks and exit.",
    )


def list_(config: ListConfig):
    console = Console()
    if config.cookbooks:
        default_cookbooks = Cookbook.load_all_defaults()
        console.print(Markdown("\n".join(f"- {cb.name}" for cb in default_cookbooks)))
    if config.maps:
        console.print(Markdown(get_all_map_list_markdown()))
    exit(0)


def main():
    parser = ArgumentParser("sbomgrader")
    subparsers = parser.add_subparsers(required=True, dest="command")
    grade_parser = subparsers.add_parser("grade")
    create_grade_parser(grade_parser)
    convert_parser = subparsers.add_parser("convert")
    create_convert_parser(convert_parser)
    list_parser = subparsers.add_parser("list")
    create_list_parser(list_parser)
    args = parser.parse_args()

    map_ = {
        "grade": (grade, GradeConfig),
        "convert": (convert, ConvertConfig),
        "list": (list_, ListConfig),
    }
    func, config_class = map_[args.command]
    return func(config_class.from_args(args))


if __name__ == "__main__":
    main()



import datetime
from typing import Any

from sbomgrader.core.definitions import FIELD_NOT_PRESENT, TIME_ISO_FORMAT_STRING


def should_remove(item: Any) -> bool:
    """Is the field considered invalid?"""
    # All kinds of empty values are considered invalid
    # But `False` is valid
    return (
        (not item and item is not False)
        or item is FIELD_NOT_PRESENT
        or item == "Field not present."
    )


def __is_prunable(item: Any) -> bool:
    return (
        isinstance(item, dict)
        or isinstance(item, list)
        or isinstance(item, datetime.datetime)
    )


def prune(struc: Any) -> Any:
    """This function cuts off invalid values from returned structures."""
    if isinstance(struc, datetime.datetime):
        # YAML automatically transforms ISO time to python datetime.
        # Datetime is not JSON-serializable
        return struc.strftime(TIME_ISO_FORMAT_STRING)
    if not __is_prunable(struc):
        return struc
    if isinstance(struc, list):
        to_prune = []
        replace_dict = {}
        for idx, item in enumerate(struc):
            if should_remove(item):
                to_prune.append(idx)

            if __is_prunable(item):
                pruned = prune(item)
                if should_remove(pruned):
                    to_prune.append(idx)
                else:
                    replace_dict[idx] = pruned

        for idx, item in replace_dict.items():
            struc[idx] = item
        for idx in reversed(to_prune):
            struc.pop(idx)

    elif isinstance(struc, dict):
        to_prune = set()
        update_dict = {}
        for key, val in struc.items():
            if should_remove(val):
                to_prune.add(key)
            if __is_prunable(val):
                pruned = prune(val)
                if should_remove(pruned):
                    to_prune.add(key)
                else:
                    update_dict[key] = pruned
        struc.update(update_dict)
        for item in to_prune:
            struc.pop(item)
    return struc



from typing import Any

from sbomgrader.core.field_resolve import FieldResolver, Variable
from sbomgrader.translate.prune import prune
from sbomgrader.translation_maps.transformers.red_hat_spdx23_cdx16.spdx23 import (
    spdxid_to_bom_ref,
)


def deduplicate_srpm_midstreams(
    old_doc: dict[str, Any], new_doc: dict[str, Any]
) -> None:
    # First map the relationship correctly according to the old doc
    field_resolver = FieldResolver({})
    old_srpms_spdxids = field_resolver.get_objects(
        old_doc,
        "packages[externalRefs[referenceType=purl]"
        "referenceLocator%=pkg:rpm/,externalRefs[referenceType=purl]"
        "referenceLocator%arch=src]SPDXID",
    )
    midstream_bom_ref_map = {}

    for srpm_spdxid in old_srpms_spdxids:
        related_midstream_spdxids = field_resolver.get_objects(
            old_doc,
            f"relationships[spdxElementId={srpm_spdxid},"
            f"relationshipType=CONTAINS]relatedSpdxElement",
        )
        srpm_purls = field_resolver.get_objects(
            old_doc,
            f"packages[SPDXID={srpm_spdxid}]externalRefs[referenceType=purl]referenceLocator",
        )
        srpm_bom_ref = spdxid_to_bom_ref(srpm_spdxid, srpm_purls)
        midstream_bom_ref_map[srpm_bom_ref] = set()
        for related_midstream_spdxid in related_midstream_spdxids:
            related_midstream_purls = field_resolver.get_objects(
                old_doc,
                f"packages[SPDXID={related_midstream_spdxid}]externalRefs[referenceType=purl]referenceLocator",
            )
            midstream_bom_ref = spdxid_to_bom_ref(
                related_midstream_spdxid, related_midstream_purls
            )
            midstream_bom_ref_map[srpm_bom_ref].add(midstream_bom_ref)

    # Now perform the deduplication
    for component in new_doc.get("components", []):
        if not component.get("bom-ref") in midstream_bom_ref_map:
            continue
        to_remove = []
        ancestors = component.get("pedigree", {}).get("ancestors", [])
        for idx, ancestor in enumerate(ancestors):
            if (
                ancestor.get("bom-ref")
                not in midstream_bom_ref_map[component.get("bom-ref")]
            ):
                to_remove.append(ancestor.get("bom-ref"))
        for idx in reversed(to_remove):
            ancestors.pop(idx)
    prune(new_doc)


def deduplicate_srpm_upstreams(
    old_doc: dict[str, Any], new_doc: dict[str, Any]
) -> None:
    old_field_resolver = FieldResolver(
        {
            "srpm_packages_spdxids": Variable(
                "srpm_packages_spdxids",
                "packages[externalRefs[referenceType=purl]referenceLocator%=pkg:rpm/,externalRefs[referenceType=purl]referenceLocator%arch=src]SPDXID",
            )
        }
    )
    upstream_bom_ref_map = {}
    old_midstream_spdxids = old_field_resolver.get_objects(
        old_doc,
        "relationships[relationshipType=CONTAINS,spdxElementId=${srpm_packages_spdxids}]relatedSpdxElement",
    )
    for midstream_spdxid in old_midstream_spdxids:
        midstream_purls = old_field_resolver.get_objects(
            old_doc,
            f"packages[SPDXID={midstream_spdxid}]externalRefs[referenceType=purl]referenceLocator",
        )
        midstream_bom_ref = spdxid_to_bom_ref(midstream_spdxid, midstream_purls)
        upstream_bom_ref_map[midstream_bom_ref] = set()
        related_upstream_spdxids = old_field_resolver.get_objects(
            old_doc,
            f"relationships[relationshipType=GENERATED_FROM,spdxElementId={midstream_spdxid}]relatedSpdxElement",
        )
        for related_upstream_spdxid in related_upstream_spdxids:
            related_upstream_purls = old_field_resolver.get_objects(
                old_doc,
                f"packages[SPDXID={related_upstream_spdxid}]externalRefs[referenceType=purl]referenceLocator",
            )
            upstream_bom_ref = spdxid_to_bom_ref(
                related_upstream_spdxid, related_upstream_purls
            )
            upstream_bom_ref_map[midstream_bom_ref].add(upstream_bom_ref)
    # Deduplication
    new_field_resolver = FieldResolver({})
    for midstream in new_field_resolver.get_objects(
        new_doc, "components[|].?.pedigree.ancestors[|]"
    ):
        if midstream.get("bom-ref" not in upstream_bom_ref_map):
            continue
        to_remove = []
        ancestors = midstream.get("pedigree", {}).get("ancestors", [])
        for idx, ancestor in enumerate(ancestors):
            if (
                ancestor.get("bom-ref")
                not in upstream_bom_ref_map[midstream.get("bom-ref")]
            ):
                to_remove.append(idx)
        for idx in reversed(to_remove):
            ancestors.pop(idx)


def merge_dependencies(_, new_doc: dict[str, Any]) -> None:
    if "dependencies" not in new_doc:
        return
    new_dependencies = []
    for dep in new_doc.get("dependencies", []):
        dep_ref = dep["ref"]
        dep_obj = next(
            filter(lambda x: x.get("ref") == dep_ref, new_dependencies), None
        )
        if not dep_obj:
            dep_obj = {"ref": dep_ref}
            new_dependencies.append(dep_obj)
        for key in "provides", "dependsOn":
            item_list = dep.get(key)
            if item_list and key not in dep_obj:
                dep_obj[key] = []
            if item_list:
                dep_obj[key].extend(item_list)
    new_doc["dependencies"] = new_dependencies


def resolve_types(_, new_doc: dict[str, Any]) -> None:
    resolver = FieldResolver({})
    for component in resolver.get_objects(new_doc, "components[|]", {}):
        if "type" in component:
            continue
        cpes = resolver.get_objects(
            component, "?.evidence.identity[field=cpe]concludedValue"
        )
        if cpes:
            component["type"] = "operating-system"
        else:
            # Some reasonable fallback
            component["type"] = "data"


def try_to_restore_org_urls(_, new_doc: dict[str, Any]) -> None:
    main_supp = new_doc.get("metadata", {}).get("supplier", {})
    main_supp_name = main_supp.get("name")
    main_supp_url = main_supp.get("url")
    if not main_supp_name or not main_supp_url:
        return
    for component in new_doc.get("components", []):
        supp = component.get("supplier")
        if supp and supp.get("name") == main_supp_name:
            supp["url"] = main_supp_url


def clone_main_component(_, new_doc: dict[str, Any]) -> None:
    resolver = FieldResolver({})
    main_bom_ref = next(
        iter(resolver.get_objects(new_doc, "metadata.component.bom-ref")), None
    )
    if not main_bom_ref:
        return
    main_component = next(
        iter(resolver.get_objects(new_doc, f"components[bom-ref={main_bom_ref}]")), None
    )
    if main_component:
        new_doc["metadata"]["component"].update(main_component)



import enum
from pathlib import Path
from typing import Any

from sbomgrader.core.definitions import FORMAT_VALIDATION_SCHEMA_PATH, FORMAT_FILE_PATH
from sbomgrader.core.utils import get_mapping


def _load_formats_file(
    path: str | Path,
) -> tuple[
    type[enum.Enum], dict[enum.Enum, dict[str, Any]], dict[enum.Enum, set[enum.Enum]]
]:
    format_dict = get_mapping(path, FORMAT_VALIDATION_SCHEMA_PATH)
    enum_dict = {}
    expected_fields_dict = {}
    fallback_dict = {}
    for format_def in format_dict["formats"]:
        name = format_def["name"]
        value = format_def["value"]
        enum_dict[name] = value

    ans_enum = enum.Enum("Formats", enum_dict)

    for format_def in format_dict["formats"]:
        name = format_def["name"]
        expected_structure = format_def["expectedStructure"]
        fallback = format_def.get("fallback", [])

        enum_instance = getattr(ans_enum, name)
        expected_fields_dict[enum_instance] = expected_structure
        fallback_dict[enum_instance] = {getattr(ans_enum, form) for form in fallback}
    return ans_enum, expected_fields_dict, fallback_dict


SBOMFormat, SBOM_FORMAT_DEFINITION_MAPPING, SBOM_FORMAT_FALLBACK = _load_formats_file(
    FORMAT_FILE_PATH
)


def get_fallbacks(format_: SBOMFormat) -> set[SBOMFormat]:
    """Get formats that are considered a fallback for the provided format."""
    return SBOM_FORMAT_FALLBACK[format_]
